# coding: utf8
""" Attempt at gevent-compatible requests without any monkeypatching
(using geventhttpclient).


== Usage example ==

    # ### Basic imports / inits
    # # The wrapper that uses the session (or just use `get_session().request(…)`)
    from pyaux.base import request
    import gevent
    # # Pool (concurrency limiting)
    from gevent.pool import Pool
    pool = Pool(10)
    results = {}


=== Usage example: callback ===

    # # Need an extra wrapper that would make the globals of the
    # # callback itself (as the loop's globals change)
    def mk_callback(url):
        def callback(resp):
            results[url] = resp

        return callback

    # # Generalised way:
    def mk_callback_(__func, **env):
        def callback(*ar, **kwa):
            env.update(kwa)
            return __func(*ar, **env)

    for url in urls:
        pool.apply_async(
            request,
            args=(url,),
            kwds=dict(_return_unjson=True),
            callback=mk_callback(url),
            # # Generalised way:
            # callback = mk_callback_(
            #     (lambda resp, url: extras.__setitem__(url, resp)),
            #     url=url)
        )

    gevent.sleep()  # Uncertain
    pool.join(timeout=10)

    return results


=== Usage example: promises (greenlets) ===

    for url in urls:
        resp = pool.apply_async(
            request,
            args=(url,),
            kwds=dict(_return_unjson=True),
            callback=mk_callback(url)
        )
        results[url] = resp

    pool.join(timeout=10)

    results = {key: value.get() for key, value in results.items()}
    return results

"""

import geventhttpclient
from geventhttpclient import httplib as gehttplib
from geventhttpclient.useragent import CompatResponse

import httplib as _httplib

import requests
from requests import adapters
from requests.packages.urllib3.poolmanager import PoolManager as _PoolManager
from requests.packages.urllib3.poolmanager import SSL_KEYWORDS
# from requests.packages.urllib3.response import HTTPResponse
# from requests.packages.urllib3.exceptions import ...

from requests.packages.urllib3.connectionpool import HTTPConnectionPool, HTTPSConnectionPool
from requests.packages.urllib3.connection import HTTPConnection as _HTTPConnection
from requests.packages.urllib3.connection import HTTPSConnection as _HTTPSConnection

from .base import memoize


class GHTTPAdapter(adapters.HTTPAdapter):
    """ HTTP Adapter with overrides to use geventhttpclient """

    def init_poolmanager(self, connections, maxsize, block=adapters.DEFAULT_POOLBLOCK):
        super(GHTTPAdapter, self).init_poolmanager(connections, maxsize, block=block)
        self.poolmanager = PoolManager(
            num_pools=connections, maxsize=maxsize, block=block)


class PoolManager(_PoolManager):

    # XXX: copypasted (for the different globals)
    # TODO?: just hack the globals... somehow.
    def _new_pool(self, scheme, host, port):
        pool_cls = pool_classes_by_scheme[scheme]
        kwargs = self.connection_pool_kw
        if scheme == 'http':
            kwargs = self.connection_pool_kw.copy()
            for kw in SSL_KEYWORDS:
                kwargs.pop(kw, None)

        return pool_cls(host, port, **kwargs)


class HTTPResponseProxy(object):

    def __init__(self, resp):
        self.__resp = resp
        # NOTE: Making this class a second layer of compatibility (out
        # of 4 total, generally)
        self.__respc = CompatResponse(resp)
        self.msg = HTTPMessageAlike(resp.headers)
        self.status = int(resp.status)
        # self.__content_consumed = False

    # @property
    # def fp(self):
    #     """ A hack for
    #     `requests.packages.urllib3.util.response.is_fp_closed`
    #     which is used in
    #     `requests.packages.urllib3.response.HTTPResponse.stream`
    #     """
    #     # if self.__content_consumed:
    #     #     return None
    #     if self._body_buffer:
    #         # Not everything from the geventhttpclient's buffer was read
    #         return self
    #     return None  # 'virtually closed'
    # def read(self, *ar, **kwa):
    #     res = self.__respc.read(*ar, **kwa)
    #     if not res:
    #         self.__content_consumed = True
    #     return res

    # @property
    # def closed(self):
    #     return self.__resp.isclosed()

    @property
    def closed(self):
        return not self._body_buffer

    def __getattr__(self, name):
        try:
            return getattr(self.__respc, name)
        except AttributeError:
            try:
                return getattr(self.__resp, name)
            except AttributeError:
                if name == 'length':
                    return None
                raise


class HTTPMessageAlike(_httplib.HTTPMessage, object):
    def __init__(self, data):
        self.dict = dict(data)
        self.headers = ['%s: %s\r\n' % (k, v) for k, v in self.dict.items()]


fix_gehttplib_httpresponse = HTTPResponseProxy


def _getresponse(self, *ar, **kwa):
    """ Override-method for Connection that wraps the response for
    compatibility """
    # res = super(HTTPConnection, self).getresponse(*ar, **kwa)
    # res = gehttplib.HTTPConnection.getresponse(self, *ar, **kwa)
    res = self._upcls.getresponse(self, *ar, **kwa)
    res = fix_gehttplib_httpresponse(res)
    return res


def _urlopen(self, *ar, **kwa):
    """ HAX: force decode_content=True to make gzip work with the
    resulting session. Not sure what should be the correct way to do
    this.

    Applied to ConnectionPool classes.
    """
    kwa['decode_content'] = True
    return self._upcls.urlopen(self, *ar, **kwa)


class HTTPConnection(gehttplib.HTTPConnection):
    # Compat, needed for requests >= 2.2
    is_verified = False
    # Working around the old-style classes DRYishly:
    _upcls = gehttplib.HTTPConnection
    getresponse = _getresponse


class HTTPSConnection(gehttplib.HTTPSConnection):
    """ ... """
    is_verified = False
    _upcls = gehttplib.HTTPSConnection
    getresponse = _getresponse


#     def connect(self):
#         res = super(HTTPConnection, self).connect()
#         self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY,
#                         self.tcp_nodelay)
#         return res


class GHTTPConnectionPool(HTTPConnectionPool):
    ConnectionCls = HTTPConnection
    _upcls = HTTPConnectionPool
    urlopen = _urlopen


class GHTTPSConnectionPool(HTTPSConnectionPool):
    ConnectionCls = HTTPSConnection
    _upcls = HTTPSConnectionPool
    urlopen = _urlopen


pool_classes_by_scheme = {
    'http': GHTTPConnectionPool,
    'https': GHTTPSConnectionPool,
}


def make_session():
    session = requests.Session()
    session.mount('http://', GHTTPAdapter())
    session.mount('https://', GHTTPAdapter())
    return session


Session = make_session


@memoize
def get_cached_session(*ar, **kwa):
    return make_session(*ar, **kwa)


def get_session(new=False, **kwa):
    return make_session(**kwa) if new else get_cached_session(**kwa)



# coding: utf8
"""
Sting interpolation functions.

For convenience, not performance.

Generally not particularly recommended.
"""

from __future__ import print_function, absolute_import, division

import sys
import re


__all__ = (
    'interp',
    'edi',
    'InterpolationEvaluationException',
)


# http://rightfootin.blogspot.com/2007/02/string-interpolation-in-python.html
def interp(string, _regexp=r'(#\{([^}]*)\})'):
    """ Inline string interpolation.
    >>> var1 = 213; ff = lambda v: v**2
    >>> interp("var1 is #{var1}")
    'var1 is 213'
    >>> interp("var1 is #{ff(var1)}; also #{ff(12)}")
    'var1 is 45369; also 144'
    """
    fframe = sys._getframe(1)
    flocals = fframe.f_locals
    fglobals = fframe.f_globals
    items = re.findall(_regexp, string)
    item_to_str = {}
    # Do eval and replacement separately and replacement in one regex
    # go to avoid interpolating already interpolated values.
    for item_outer, item in items:
        item_to_str[item] = str(eval(item, fglobals, flocals))
    string = re.sub(_regexp, lambda match: item_to_str[match.group(2)], string)
    return string


class InterpolationEvaluationException(KeyError):
    pass


class edi(dict):  # "expression_dictionary"...
    """ Yet another string interpolation helper.

    >>> var1 = 313; f = lambda x: x*2
    >>> print("1 is %(var1)05d, f1 is %(f(var1))d, f is %(f)r, 1/2 is %(float(var1)/2)5.3f." % edi())  #doctest: +ELLIPSIS
    1 is 00313, f1 is 626, f is <function <lambda> at 0x...>, 1/2 is 156.500.

    """
    # No idea for what sake this is subclassed from dictionary, actually. A
    # neat extra, perhaps.

    globals = {}

    def __init__(self, d=None):
        if d is None:  # Grab parent's locals forcible
            self.locals = sys._getframe(1).f_locals
            self.globals = sys._getframe(1).f_globals
            d = self.locals
        super(edi, self).__init__(d)

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            try:
                return eval(key, self.globals, self)
            except Exception as exc:
                raise InterpolationEvaluationException(key, exc)



# coding: utf8
"""
A collection of useful helpers.
 """

__version__ = "2.9.0"

# Note: no modules imported here should import `decimal` (otherwise
# `use_cdecimal` might become problematic for them)

from .base import *

from . import ranges
from .ranges import *

from . import interpolate
from .interpolate import *


__all__ = (
    'bubble',
    'window',
    'dotdict',
    'SmartDict',
    'DebugPlug', 'repr_call',
    'dict_fget',
    'dict_fsetdefault',
    'split_list',
    'use_cdecimal',
    'use_exc_ipdb',
    'use_exc_log',
    'use_colorer',
    'obj2dict',
    'mk_logging_property',
    'sign',
    'try_parse',
    'human_sort_key',
    'reversed_blocks',
    'reversed_lines',
    'lazystr',
    'list_uniq',
    'o_repr',
    'chunks',
    'chunks_g',
    # 'runlib',
    # 'lzmah',
    # 'lzcat',
    # 'psql',
    'to_bytes',
    'to_unicode',
) + ranges.__all__ + interpolate.__all__



# coding: utf8
""" madstuff: other stuff stuff """

import re
from six.moves import urllib_parse as urlparse
from six.moves import xrange
from pyaux import dotdict
from pyaux.dicts import MVOD
from pyaux.base import repr_cut as _cut


__all__ = (
    'Url',
    '_url_re',
    '_cut', 'IPNBDFDisplay',
    '_re_largest_matching_start',
)


# See also: https://pypi.org/project/yarl/
class Url(dotdict):
    """ urlparse.ParseResult and parse_qs[l] in a dict-like non-lazy form """
    _components = (
        'scheme', 'netloc', 'path', 'params', 'query', 'fragment',
        # from ResultMixin
        'username', 'password', 'hostname', 'port',
    )  # <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
    _base_components = (
        'scheme', 'netloc', 'path', 'params', 'query', 'fragment',
    )

    # TODO: urlunescaped parts

    def __init__(self, url, **kwa):
        self.url = url
        urldata = urlparse.urlparse(url, **kwa)
        for key in self._components:
            val = getattr(urldata, key)
            setattr(self, key, val)

        self.query_str = urldata.query
        self.queryl = urlparse.parse_qs(urldata.query)
        self.query = MVOD(urlparse.parse_qsl(urldata.query))
        # TODO?: self.query = pyaux.dicts.MVOD(urldata.query)

    def to_string(self):
        query_str = urlparse.urlencode(list(self.query.items()))
        return urlparse.urlunparse((
            self[key] if key != 'query' else query_str
            for key in self._base_components))


_url_re = (
    r'''(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+'''
    r'''[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+'''
    r'''(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’]))''')


def IPNBDFDisplay(df, *ar, **kwa):
    """ A helper to display a pandas DataFrame in the IPython notebook.

    Recommended things to do:
    pandas.options.display.max_colwidth = 2000
    H = IPNBDFDisplay

    :param head: (default 200) runs df = df.head(head); has default to
        avoid accidentally outputing too much; needs explicit
        `head=None` to disable.

    :param cutlinks: automatically converts all URLs in the resulting HTML
    to shortened values leading to the same addresses.
    """
    from IPython.display import display, HTML
    kwa.setdefault('float_format', lambda v: '%.6f' % (v,))
    columns = kwa.pop('columns', None)
    include = kwa.pop('include', None)
    exclude = kwa.pop('exclude', None)
    head = kwa.pop('head', 200)
    cutlinks = kwa.pop('cutlinks', True)
    do_display = kwa.pop('display', True)

    if head:
        df = df.head(head)

    if columns is not None:  # allows for reordering
        df = df.__class__(df, columns=columns)
    if include is not None:
        df = df.__class__(df, columns=[val for val in df.columns if val in include])
    if exclude is not None:
        df = df.__class__(df, columns=[val for val in df.columns if val not in exclude])

    html = df.to_html(*ar, **kwa)

    if cutlinks:
        cutlinks = 80 if cutlinks is True else cutlinks

        def cutlink(link):
            result = '<a href="%s">%s</a>' % (link, _cut(link, cutlinks))
            return result

        html = re.sub(
            _url_re,
            lambda match: cutlink(match.group(0)),
            html)

    # TODO?: option to insert copious '<wb/>'s in all cells

    result = HTML(html)
    if do_display:
        display(result)
    return result


def _re_largest_matching_start(regex, value, return_regexp=False):
    """ Find a largest match (from the start of the string) in a value
    for the regex.

    WARN: computationally complex (d'uh).

    >>> _re_largest_matching_start(r'^[az]+zxcvb', 'aazx')
    'aazx'
    >>> _re_largest_matching_start(r'^[az]+zxcvb', 'aazx', return_regexp=1)
    ('^[az]+zx', 'aazx')
    """
    # Yet Another Insane Horror

    all_regexes = [regex[:idx] for idx in xrange(len(regex) + 1)]

    def _try_match(rex, st):
        try:
            return re.match(rex, st)
        except Exception:
            return

    # all_match_tries = [_try_match(subreg, s) for subreg in all_regexes]

    # Even more horrible:
    all_substrings = [value[:idx] for idx in xrange(len(value) + 1)]
    all_match_tries = (
        (subreg, _try_match(subreg, substr))
        for subreg in all_regexes
        for substr in all_substrings)
    all_matchstrings = [
        (subreg, val.group(0))
        for subreg, val in all_match_tries
        if val]
    if not all_matchstrings:
        return ''
    lrex, lval = max(
        all_matchstrings,
        key=lambda val: len(val[1]))  # longest match
    if return_regexp:
        return lrex, lval
    return lval



#!/usr/bin/env python
# coding: utf8

from __future__ import print_function, unicode_literals, absolute_import, division

import os
import json
import sys

from pyaux.base import to_bytes


def main():
    indent = int(os.environ.get('INDENT') or '2')
    if len(sys.argv) >= 2:
        with open(sys.argv[1]) as fo:
            data_in = fo.read()
    else:
        data_in = sys.stdin.read()

    try:
        outbuf = sys.stdout.buffer
    except AttributeError:
        outbuf = sys.stdout

    def bailout(msg):
        sys.stderr.write(
            "ERROR: fjson.py: %s; original data as follows (on stdout)\n" % (msg,))
        outbuf.write(to_bytes(data_in))
        return 13

    try:
        data = json.loads(data_in)
    except Exception as exc:
        return bailout("Error parsing as json: %s" % (exc,))

    try:
        data_out = json.dumps(
            data, indent=indent, sort_keys=True, ensure_ascii=False)
    except Exception as exc:
        return bailout("Error dumping as json: %s" % (exc,))

    outbuf.write(to_bytes(data_out))
    outbuf.write(b"\n")
    outbuf.flush()


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
""" lzma (pylzma) helpers.
Also can be used as a script for compressing a file.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import sys
import pylzma


def lzma_compress(fi, fo, fi_close=True, fo_close=True, bufs=65535):
    """ Compress `fi` into `fo` (`file` or filename) """
    if isinstance(fi, str):
        fi, fi_n = open(fi, 'rb'), fi
        #fi_close = True
    if isinstance(fo, str):
        fo, fo_n = open(fo, 'wb'), fo
        #fo_close = True
    #fi.seek(0)
    s = pylzma.compressfile(fi)
    while True:
        tmp = s.read(bufs)
        if not tmp:
            break
        fo.write(tmp)
    if fo_close:
        fo.close()
    if fi_close:
        fi.close()
    return fi, fo


class _IgnoreTheError(Exception):
    """ Used in `unjsllzma` to signify that the exception should be simply ignored
    """


def _handle_fail_default(v, e):
    raise _IgnoreTheError
    # supposedly can do simple `raise` to re-raise the original (parse) exception


def unjsllzma(fi, fi_close=True, parse_fn=None, handle_fail=None, bufs=655350):
    """ Make a generator for reading an lzma-compressed file with
    json(or something else) in lines.
    `parse_fn` is th function(v) to process lines with (defaults to
      `json.loads`)
    `handle_fail` if a fuction(value, exception) for handling a failure to
    parse the value; value is skipped if it raises _IgnoreTheError
    exception, otherwise its return value is yielded.  default: skip all
    failures.
    """
    if parse_fn is None:
        try:
            import simplejson as json
        except ImportError:
            print("Error importing (preferred) simplejson")
            import json
        parse_fn = json.loads

    if handle_fail is None:
        handle_fail = _handle_fail_default

    def try_loads(v):
        try:
            return parse_fn(v)
        except Exception as e:
            return handle_fail(v, e)

    if isinstance(fi, str):
        fi = open(fi, 'rb')

    tmp2 = ''  # buffer for unfunushed lines
    in_bufs = int(bufs / 100)  # XXX: see lzcat.py note around in_bufs
    s = pylzma.decompressobj()
    cont = True
    while cont:
        tmp = fi.read(in_bufs)
        if not tmp:  # nothing more can be read
            tmp2 += s.flush()
            cont = False
        else:
            # XXX: TODO: use bytearray.extend (likely).
            tmp2 = tmp2 + s.decompress(tmp, bufs)
        tmp3 = tmp2.split('\n')  # finished and unfinished lines
        for v in tmp3[:-1]:
            try:
                r = try_loads(v)
            except _IgnoreTheError:
                continue  # no more handling requested, just skip it
            yield r
        tmp2 = tmp3[-1]
    if fi_close:
        fi.close()


def get_stdin():
    """
    Get a stdin fileobject that can (possibly) read bytes.
    """
    try:
        return sys.stdin.buffer  # py3
    except AttributeError:
        return sys.stdin  # py2


def get_stdout():
    """
    Get stdout fileobject that can possibly read bytes.
    """
    try:
        return sys.stdout.buffer  # py3
    except AttributeError:
        return sys.stdout  # py2


def _lzma_main():
    fi_a = sys.argv[1]
    fo_a = sys.argv[2]

    if fi_a == '-':
        fi_a = get_stdin()
    if fo_a == '-':
        fo_a = get_stdout()
    lzma_compress(fi_a, fo_a)


if __name__ == '__main__':
    _lzma_main()



# coding: utf8
"""
Various range-generating funcs.
"""

from __future__ import absolute_import, print_function



__all__ = (
    'fxrange', 'frange', 'dxrange', 'drange',
    'date_xrange', 'date_range',
    'date_add_months', 'date_months_xrange', 'date_months_range',
)


def fxrange(start, end=None, inc=None):
    """ The xrange function for float """
    assert inc != 0, "inc should not be zero"
    if end is None:
        end = start
        start = 0.0
    if inc is None:
        inc = 1.0
    i = 0  # to prevent error accumulation
    while True:
        nextv = start + i * inc
        if (inc > 0 and nextv >= end or
                inc < 0 and nextv <= end):
            break
        yield nextv
        i += 1


def frange(start, end=None, inc=None):
    """ list(fxrange) """
    return list(fxrange(start, end, inc))


def dxrange(start, end=None, inc=None, include_end=False):
    """ The xrange function for Decimal """
    # Imported here mostly because of use_cdecimal in this module
    from decimal import Decimal
    assert inc != 0, "inc should not be zero"
    if end is None:
        end = start
        start = 0
    if inc is None:
        inc = 1
    inc = Decimal(inc)
    start = Decimal(start)
    end = Decimal(end)
    nextv = start
    while True:
        if ((inc > 0) and (not include_end and nextv == end or nextv > end) or
                (inc < 0) and (not include_end and nextv == end or nextv < end)):
            break
        yield nextv
        nextv += inc


def drange(*ar, **kwa):
    """ list(dxrange) """
    return list(dxrange(*ar, **kwa))


def date_xrange(start, end, inc=None, include_end=False, precise=False):
    """ The xrange function for datetime.

    NOTE: the semantics of 'start' and 'end' are different here: with
    end=None an infinite generator is returned.

    :param precise: do more calculations but potentially produce more
        precise results (especially for precise `inc`).

    >>> import datetime
    >>> dt = datetime.datetime
    >>> dta = dt(2011, 11, 11)
    >>> dtb = dt(2011, 11, 14)
    >>> dsl = lambda dtl: [dt.strftime('%Y-%m-%d') for dt in dtl]
    >>> dtsl = lambda dtl: [dt.isoformat() for dt in dtl]
    >>> dsl(date_xrange(dta, dt(2011, 11, 13)))
    ['2011-11-11', '2011-11-12']
    >>> dsl(date_xrange(dta, dt(2011, 11, 13), include_end=True))
    ['2011-11-11', '2011-11-12', '2011-11-13']
    >>> dsl(date_xrange(dta, dtb, inc=2))
    ['2011-11-11', '2011-11-13']
    >>> dtsl(date_xrange(dta, dtb, inc=1111.11111111111 / 86400, precise=True))[-1]
    '2011-11-13T23:54:48.888889'
    >>> dtsl(date_xrange(dta, dtb, inc=1111.11111111111 / 86400))[-1]
    '2011-11-13T23:54:48.888863'
    """
    import datetime
    if inc is None:
        inc = 1  # default: 1 day

    if not isinstance(inc, datetime.timedelta):
        # NOTE: days, by default
        inc_days = inc
        inc = datetime.timedelta(inc)
    else:
        inc_days = inc.total_seconds() / 86400  # py2.7 required

    assert inc_days  # should be nonzero

    is_forward = (inc_days > 0)

    idx = 0
    current = start
    while True:
        if end is not None:
            if include_end:
                to_break = current > end if is_forward else current < end
            else:
                to_break = current >= end if is_forward else current <= end
            if to_break:
                break
        yield current
        if precise:
            idx += 1
            current = start + datetime.timedelta(inc_days * idx)
        else:
            current = current + inc


def date_range(*ar, **kwa):
    """ list(date_xrange) """
    return list(date_xrange(*ar, **kwa))


def date_add_months(sourcedate, months=1):
    """ Add months to date; can cap the day to the maximal value for
    the month """
    import calendar
    month = sourcedate.month - 1 + months
    year = sourcedate.year + month // 12
    month = month % 12 + 1
    day = min(sourcedate.day, calendar.monthrange(year, month)[1])
    # return datetime.date(year, month, day)
    # On the other hand, we can keep the type and the time:
    return sourcedate.replace(year=year, month=month, day=day)


def date_months_xrange(start, end, inc=1, include_end=False):
    """ date_range with delta measured in months.

    Similar semantics to date_xrange.
    Only accepts integer `inc` values.

    >>> import datetime
    >>> dt = datetime.datetime
    >>> dsl = lambda dtl: [dt.strftime('%Y-%m-%d') for dt in dtl]
    >>> dsl(date_months_xrange(dt(2011, 10, 31), dt(2012, 1, 1)))
    ['2011-10-31', '2011-11-30', '2011-12-31']
    """
    assert isinstance(inc, int)
    inc = int(inc)
    assert inc  # should be nonzero
    is_forward = inc > 0
    current = start
    idx = 0
    while True:
        if end is not None:
            if include_end:
                to_break = current > end if is_forward else current < end
            else:
                to_break = current >= end if is_forward else current <= end
            if to_break:
                break
        yield current
        idx += 1
        current = date_add_months(start, months=inc * idx)


def date_months_range(*ar, **kwa):
    """ list(date_months_xrange) """
    return list(date_months_xrange(*ar, **kwa))



# coding: utf8
""" Set the exception handler to additionaly log the exception before
processing it further (e.g. to e-mail it to admins).

Logs local variables at each traceback level when possible

Replaces sys.excepthook on `init()`.
Can be included in 'sitecustomize.py'
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import re
import sys
import traceback
import logging

# # Extracts from django.views.debug
# # (should not require django)
try:  # pretty printing
    # # That one is a little bit preettier
    from IPython.lib.pretty import pretty as pformat
except ImportError:
    from pprint import pformat
# from django.template.filters import force_escape
from six.moves import reprlib
from pyaux import edi  # 'templating'.
from pyaux.base import to_text


_log = logging.getLogger("unhandled_exception_handler")

_lrepr_params = dict(
    maxlevel=8, maxtuple=64, maxlist=64, maxarray=48,
    maxdict=64, maxset=64, maxfrozenset=64, maxdeque=64, maxstring=80,
    maxlong=128, maxother=32,
)


def make_lrepr():
    lrepr = reprlib.Repr()
    for key, val in _lrepr_params.items():
        setattr(lrepr, key, val)
    return lrepr


# # The singleton:
LREPR = make_lrepr()


def lrepr_call(value):
    return LREPR.repr(value)


def info(exc_type, exc_value, tb):
    _log.exception('')
    # call the default hook
    sys.__excepthook__(exc_type, exc_value, tb)


def _var_repr(v, ll=356):
    try:
        # # not exactly optimized in case of huge datalists
        # r = pformat(v)
        # # not exactly... pretty
        r = lrepr_call(v)
        # # XXX: combine those two somehow?
        # # (also, make it print last value of `list`/`deque`/... always, too)
    except Exception as exc:
        return "<un`repr()`able variable: {!r}>".format(exc)
    # if len(r) > ll:  # handled by the lrepr, somewhat; `ll` is ignored
    #     return r[:ll-4] + '... '
    return r


def _get_lines_from_file(filename, lineno, context_lines, loader=None, module_name=None):
    """
    Returns context_lines before and after lineno from file.
    Returns (pre_context_lineno, pre_context, context_line, post_context).
    """
    source = None
    if loader is not None and hasattr(loader, "get_source"):
        source = loader.get_source(module_name)
        if source is not None:
            source = source.splitlines()
    if source is None:
        try:
            f = open(filename)
            try:
                source = f.readlines()
            finally:
                f.close()
        except (OSError, IOError):
            pass
    if source is None:
        return None, [], None, []

    encoding = 'ascii'
    for line in source[:2]:
        # File coding may be specified. Match pattern from PEP-263
        # (http://www.python.org/dev/peps/pep-0263/)
        match = re.search(r'coding[:=]\s*([-\w.]+)', line)
        if match:
            encoding = match.group(1)
            break
    source = [to_text(sline, encoding=encoding, errors='replace') for sline in source]

    lower_bound = max(0, lineno - context_lines)
    upper_bound = lineno + context_lines

    pre_context = [line.strip('\n') for line in source[lower_bound:lineno]]
    context_line = source[lineno].strip('\n')
    post_context = [line.strip('\n') for line in source[lineno+1:upper_bound]]

    return lower_bound, pre_context, context_line, post_context


def get_traceback_frame_variables(tb_frame):
    return tb_frame.f_locals.items()


def get_traceback_frames(tb):
    frames = []
    while tb is not None:
        # Support for __traceback_hide__ which is used by a few libraries
        # to hide internal frames.
        if tb.tb_frame.f_locals.get('__traceback_hide__'):
            tb = tb.tb_next
            continue
        filename = tb.tb_frame.f_code.co_filename
        function = tb.tb_frame.f_code.co_name
        lineno = tb.tb_lineno - 1
        loader = tb.tb_frame.f_globals.get('__loader__')
        module_name = tb.tb_frame.f_globals.get('__name__') or ''
        pre_context_lineno, pre_context, context_line, post_context = _get_lines_from_file(filename, lineno, 7, loader, module_name)
        if pre_context_lineno is not None:
            frames.append({
                'tb': tb,
                #'type': module_name.startswith('django.') and 'django' or 'user',
                'filename': filename,
                'function': function,
                'lineno': lineno + 1,
                'vars': get_traceback_frame_variables(tb.tb_frame),
                'id': id(tb),
                'pre_context': pre_context,
                'context_line': context_line,
                'post_context': post_context,
                'pre_context_lineno': pre_context_lineno + 1,
            })
        tb = tb.tb_next

    return frames


def _exc_safe_repr(exc_type, exc_value):
    """ Mildly paranoid extraction of repr of exception """
    ## NOTE: returns a terminating newline. Strip if a problem
    return ''.join(traceback.format_exception_only(exc_type, exc_value))
    #return (
    #  getattr(exc_type, '__module__', '<unknown module>'),
    #  getattr(exc_type, '__name__', '<unnamed exception!?>'),
    #  getattr(exc_value, 'message', '<no message!?>'),
    #  )


def render_exc_repr(exc_type, exc_value):
    """ A still paranoid but more detailed exception representator """
    res = ''
    try:
        res += "Error:  %s" % _exc_safe_repr(exc_type, exc_value)
        try:
            res += "  (repr: '''%r''')\n" % (exc_value,)
        except Exception as e:
            try:
                res += "  (Failure to repr: %r)\n" % (e,)
            except Exception as e2:
                res += "  (Failure to repr totally: (%s) (%s)\n" % (
                    _exc_safe_repr(type(e), e).strip(),
                    _exc_safe_repr(type(e2), e2).strip())
    except Exception as e3:
        try:
            res += (
                "Error: Some faulty exception of type %r, failing "
                "on repr with %s") % (
                    exc_type, _exc_safe_repr(type(e3), e3))
        except Exception:
            res += "Error: Some very faulty exception"
    return res


def render_frames_data(frames, exc_type=None, exc_value=None):
    ## A crappy way to do that compared to templates, but w/e really
    res = ""
    if exc_type or exc_value:
        ## Also convenient:
        res += _exc_safe_repr(exc_type, exc_value)
    if frames:
        res += "Traceback details:\n"
        for frame in frames:
            #res += ("%(frame['filename'])s:%(frame['lineno'])d:"
            #  " %(frame['function'])r -> %(frame['context_line'])s\n") % edi()
            res += (
                "---- File %(frame['filename'])s, line %(frame['lineno'])d, in"
                " %(frame['function'])r:\n  > %(frame['context_line'])s\n") % edi()
            # if frame.pre_context: frame.id; for i, line in enumerate(frame.pre_context): frame.pre_context_lineno + i, line
            # if frame.context_line:
            #     res += "  %(frame.lineno): %(frame.context_line)s" % edi()
            # if frame.post_context: for i, line in enumerate(frame.post_context): frame.lineno + 1 + i, line
            if frame['vars']:
                res += "  Local vars:"
                for var in sorted(frame['vars'], key=lambda v: v[0]):
                    # Note: 13 spaces to visually separate the
                    #   variables at the same time taking less vertical
                    #   space than printing each from a new line (and
                    #   then adding spaces anyway).
                    res += " " * 13
                    res += "%(var[0])s: %(var[1])s;" % edi()
                res += "\n"
    if exc_type or exc_value:
        res += render_exc_repr(exc_type, exc_value)
    return res


def advanced_info(exc_type, exc_value, tb):
    #reporter = ExceptionReporter(None, exc_type, exc_value, tb)
    frames = get_traceback_frames(tb)
    for idx, frame in enumerate(frames):
        if 'vars' in frame:
            frame['vars'] = [
                (key,
                 # force_escape(pprint(v))
                 _var_repr(val)
                ) for key, val in frame['vars']]
        frames[idx] = frame  # XX: does this even do something?

    text = render_frames_data(frames, exc_type, exc_value)
    #_log.exception(text)
    _log.error(text)
    # print(text)
    sys.__excepthook__(exc_type, exc_value, tb)


def advanced_info_safe(exc_type, exc_value, tb):
    """ A paranoid wrapper around handlers; not normally necessary as Python
    handles unhandled exceptions in unhandled exception handlers properly
    """
    try:
        advanced_info(exc_type, exc_value, tb)
    except Exception as e:
        try:
            info(exc_type, exc_value, tb)
        except Exception:
            print("Everything is failing! Running the default excepthook.")
            sys.__excepthook__(exc_type, exc_value, tb)
        raise e


def init():
    #sys.excepthook = info
    sys.excepthook = advanced_info
    #sys.excepthook = advanced_info_safe



# coding: utf8
"""
Simple `get whatever json`.

See also:
https://bitbucket.org/runeh/anyjson/src/default/anyjson/__init__.py
"""

from __future__ import division, absolute_import, print_function, unicode_literals

try:
    import anyjson
    # Implementation that might be faster but does not provide any extra arguments.
    json_loads_simple = anyjson.loads  # pylint: disable=invalid-name
    json_dumps_simple = anyjson.dumps  # pylint: disable=invalid-name
except Exception:  # pylint: disable=broad-except
    pass

try:
    import simplejson as json
except Exception:  # pylint: disable=broad-except
    import json


# Implementation that at least supports `ensure_ascii=False`.
json_loads = json.loads  # pylint: disable=invalid-name
json_dumps = json.dumps  # pylint: disable=invalid-name



#!/usr/bin/env python
# coding: utf8
"""
A script to convert convenient text to markdown to html.
"""


import os
import sys
import re
import pyaux


def repl_header(match):
    """ Replace '=' header with '#' header """
    lh, text, rh = match.groups()
    anchor = text.strip().replace(' ', '_')
    res = '%(lh)s %(text)s <a name="%(anchor)s" href="#%(anchor)s">§</a> %(rh)s' % dict(
        lh='#' * len(lh), text=text, anchor=anchor, rh='#' * len(rh))
    # # Add an anchor
    # res = '<a href="%s">\n%s\
    # res += '\n<a name="%s"></a>' % (text.strip(),)
    return res


class Worker:
    result = None
    state = None
    state__list = None
    state__location = None

    simple_replacements = (
        # header
        (r'^(=+) (.+) (=+)$', repl_header),
        # trailing whitespaces
        (r' *$', ''),
    )

    header = '''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8" />
    <style type="text/css">
      h1, h2, h3 {
        line-height: 1.2;
      }
      body {
        margin: 40px auto;
        max-width: 650px;
        line-height: 1.6;
        font-size: 18px;
        color: #444;
        padding: 0 10px;
      }
      ol {
        padding: 0 0 0 2em;
      }
      ol[manual] {
        padding-left: 1em;
        list-style: none;

      }
      ol[manual] li {
        padding-top: 0.25em;

        text-indent: -1em;
        padding-left: 1em;
      }
      li[value]:before {
        content: attr(value) ". ";
      }
    </style>
  </head>
  <body>
    '''

    footer = '''
  </body>
</html>
    '''

    list_header = '<ol manual=1>'
    list_footer = '</ol>'
    item_footer = '</li>'  # non-item-specific

    def process(self, lines):
        self.state = ''
        self.state__list = []
        self.state__location = []
        self.result = []

        # self.result.append(self.header)

        lines = (self.handle_simple_replacements(line) for line in lines)
        line_iter = pyaux.window(lines, fill_left=True, fill=None)

        for prev_line, line in line_iter:
            self.check_state(prev_line, line)
            if self.state == 'list':
                self.process_list(line)
            else:
                self.result.append(line)

        # self.result.append(self.footer)

        return self.result

    def handle_simple_replacements(self, line):
        # simple replacements
        for rex, repl in self.simple_replacements:
            line = re.sub(rex, repl, line)
        return line

    def check_state(self, prev_line, line):
        """ State changes handler """
        if self.state == 'list':
            # empty line and then a non-list.
            # Almost markdown-like behaviour for lists: single empty
            # lines will not break the list; but unlike markdown, two
            # empty lines will.
            if not prev_line and not re.search(r'^ +[0-9].*\. ', line):
                # Not a list anymore
                self.result, prev_line_x = self.result[:-1], self.result[-1]
                self.unwind_list()
                self.result.append(prev_line_x)
                self.state = ''
                self.state__list = None
        elif not prev_line and line.startswith(' 1. '):
            # note the very specific list starter
            self.state = 'list'
            self.state_list = []

        if self.state == '':
            # Process the headings-state
            self.check_state_header(line)

    def check_state_header(self, line):
        """ Keep the state__location current """
        heading_match = re.search(
            r'^(?P<tag>#+).*name="(?P<name>[^"]+)".*',
            line)
        if not heading_match:
            return

        data = heading_match.groupdict()
        depth = len(data['tag'])
        # Should almost always exist because of repl_header.
        name = data.get('name') or ''
        data.update(depth=depth, name=name)
        if not self.state__location:
            self.state__location = [data]
            return

        shallower, same_or_deeper = pyaux.split_list(
            self.state__location, lambda info: info['depth'] < depth)
        self.state__location = shallower + [data]

    def process_list(self, line):
        # any line should match
        match = re.search(
            r'^(?P<spaces> *)(?:(?P<num>[0-9a-z.]+)\. )?(?P<text>.*)$',
            line)
        data = match.groupdict()
        spaces = data['spaces']
        indent = len(spaces)
        num = data.get('num')
        text = data['text']
        # Synopsis: markdown considers everything within html tags to
        # be written as-is. Therefore, process all htat stuff
        # explicitly.
        # TODO: this all should've probably been done as markdown
        # extender subclass.
        text = _markdown_process(text)
        text = re.sub(r'^ *<p>(.*)</p> *$', r'\1', text)

        if not num:
            # put as-is
            # e.g.: empty lines
            self.result.append(line)
            return

        item_header = '<li value="%s">' % (num,)

        anchor_name = ''.join(
            "%s__" % (info['name'],) for info in self.state__location)
        anchor_name = anchor_name + num.rstrip('.').replace('.', '_')
        item_header = item_header + '<a name="%s"></a>' % (anchor_name,)
        item_footer = self.item_footer
        item_info = dict(data, indent=indent)

        # else:  if num:
        if not self.state__list:
            # starting a list
            self.result.extend((
                self.list_header,  # <ol>
                spaces + item_header,  # <li>
                spaces + text))
            self.state__list = [item_info]
            return

        # else: if within a list already:
        last_info = self.state__list[-1]
        if last_info['indent'] == indent:
            # same indent, i.e. continuing the list
            self.result.extend((
                spaces + item_footer,  # </li>
                spaces + item_header,  # <li>
                spaces + text))
            last_info.update(item_info)  # replace the num for possible recursion
        elif 0 < indent - last_info['indent'] <= 2:
            # going deeper
            # ol-li-ol-li chain
            self.result.extend((
                spaces + self.list_header,  # <ol>
                spaces + item_header,  # <li>
                spaces + text))
            self.state__list.append(item_info)
        elif indent - last_info['indent'] < 0:
            # returning
            state_closing, state_remain = pyaux.split_list(
                self.state__list, lambda info: info['indent'] > indent)
            self.unwind_list(state_closing)  # </li></ol>
            self.state__list = state_remain
            self.result.extend((
                spaces + item_footer,  # </li>
                spaces + item_header,  # <li>
                spaces + text))
        else:  # more than 2 spaces deeper in; assume it's just more text
            self.result.append(line)

    def unwind_list(self, infos=None):
        if infos is None:
            infos = self.state__list or []
        for info in reversed(infos):
            spaces = info['spaces']
            self.result.extend((
                spaces + self.item_footer,  # </li>
                spaces + self.list_footer))  # </ol>


def _markdown_process(text, *ar, **kwa):
    import markdown
    # Not exactly by the standard, but visually better for the lists:
    kwa.setdefault('tab_length', 2)
    return markdown.Markdown(*ar, **kwa).convert(text)


def main(src_ext='.txt'):
    if len(sys.argv) > 1:
        filename = sys.argv[1]
    else:
        filename = os.path.join(os.path.dirname(__file__), '..', 'doc.txt')

    basename = filename
    if basename.endswith(src_ext):
        basename = basename[:-len(src_ext)]

    with open(filename) as fo:
        data = fo.read()


    lines = data.splitlines()

    worker = Worker()
    result = worker.process(lines)
    result_s = '\n'.join(result)

    if not os.environ.get('NO_MD'):
        with open(basename + '.md', 'w') as fo:
            fo.write(result_s)

    import markdown
    result_html_base = _markdown_process(result_s)
    result_html = worker.header + result_html_base + worker.footer
    with open(basename + '.html', 'w') as fo:
        fo.write(result_html)


if __name__ == '__main__':
    main()



# coding: utf8
"""
Various dict-related special classes.

>>> mod = remvdodd([(1, 1), (2, 1.4), (1, 2)])
>>> mod
remvdodd(1: 1, 2: 1.4, 1: 2)
>>> mod['mod'] = mod
>>> mod
remvdodd(1: 1, 2: 1.4, 1: 2, 'mod': ...)
>>> mod.modmod = mod['mod']
>>> mod
remvdodd(1: 1, 2: 1.4, 1: 2, 'mod': ..., 'modmod': ...)
>>> import copy
>>> mc = copy.deepcopy(mod)
>>> mod.mod is mod.mod, mc.mod is not mod.mod
(True, True)
>>> mcc = mod.copy()
>>> mcc.mod is mod.mod
True
>>> mcc.clear()
>>> mcc
remvdodd()
>>> mod
remvdodd(1: 1, 2: 1.4, 1: 2, 'mod': ..., 'modmod': ...)
>>> mod.mod is mod
True
>>> mod.update(u1='y')
>>> mod
remvdodd(1: 1, 2: 1.4, 1: 2, 'mod': ..., 'modmod': ..., 'u1': 'y')
>>> mod.update_inplace([(2, 2.6), (3, 3.6)])
>>> mod
remvdodd(1: 1, 2: 2.6, 1: 2, 'mod': ..., 'modmod': ..., 'u1': 'y', 3: 3.6)
>>> mod.pop(3)
3.6
>>> mod.update_replace([(1, -1)])
>>> mod
remvdodd(2: 2.6, 'mod': ..., 'modmod': ..., 'u1': 'y', 1: -1)
>>> del mod['modmod']
>>> mod
remvdodd(2: 2.6, 'mod': ..., 'u1': 'y', 1: -1)
>>> mod._data
((2, 2.6), ('mod', remvdodd(2: 2.6, 'mod': ..., 'u1': 'y', 1: -1)), ('u1', 'y'), (1, -1))
>>> mod._data is not mod.items(), mod._data == tuple(mod.items())
(True, True)
>>> mod[1] = 3
>>> mod
remvdodd(2: 2.6, 'mod': ..., 'u1': 'y', 1: -1, 1: 3)
>>> mod.update(u1='n')
>>> mod
remvdodd(2: 2.6, 'mod': ..., 'u1': 'y', 1: -1, 1: 3, 'u1': 'n')
>>> mod
remvdodd(2: 2.6, 'mod': ..., 'u1': 'y', 1: -1, 1: 3, 'u1': 'n')
>>> list(mod.lists())
[(2, [2.6]), ('mod', [remvdodd(2: 2.6, 'mod': ..., 'u1': 'y', 1: -1, 1: 3, 'u1': 'n')]), ('u1', ['y', 'n']), (1, [-1, 3])]
>>> mod.popitem()
('u1', 'n')
>>> mod.popitem(last=False)
(2, 2.6)
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: -1, 1: 3)
>>> mod.deduplicate()
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: 3)
>>> mod.u1 = 'yy'
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: 3, 'u1': 'yy')
>>> mod.deduplicate(how='first')
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: 3)
>>> list(reversed(mod))
[1, 'u1', 'mod']
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: 3)
>>> mod.setdefault(5, 2)
2
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: 3, 5: 2)
>>> mod.setdefault(5, 1)
2
>>> mod
remvdodd('mod': ..., 'u1': 'y', 1: 3, 5: 2)
"""
# not tested:  [iter](keys|values|lists), pop, fromkeys, eq
# ipy to doctest:  sed -r 's/^In[^:]+: />>> /; s/^Out[^:]+: //; /^ *$/ d'
# test to locals:  from pyaux import dicts; reload(dicts); from pyaux.dicts import *; exec '\n'.join(v[4:] for v in dicts.__doc__.split('\n\n')[-1].splitlines() if v.startswith('>>> '))

from __future__ import print_function

import six
import copy
import itertools
try:
    from UserDict import DictMixin
except ImportError:
    from collections import MutableMapping as DictMixin

from pyaux.base import uniq_g
from pyaux.base import dotdict


__all__ = (
    'ODReprMixin', 'OrderedDict',
    'MVOD',
    'hasattr_x',
    'dotdictx', 'defaultdictx', 'DefaultDotDictMixin',
    'dodd', 'mvdodd', 'remvdodd', 'redodd',
)


class ODReprMixin(object):
    """ A mixin for ordered dicts that provides two different representations
    and a wrapper for handling self-referencing structures. """

    def __irepr__(self):
        """ The usual (default) representation of an ordereddict """
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, self.items())

    def __drepr__(self):
        """ A slightly more visual-oriented representation of an ordereddict """
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join(
                '%r: %r' % (key, val)
                for key, val in self.items()))

    def __repr__(self, _repr_running={}, fn=__drepr__):
        """ Wrapped around __drepr__ that makes it possible to
        represent infinitely-recursive dictionaries of this type. """
        # NOTE: version variety; might be a _get_ident or something else.
        try:
            from thread import get_ident as _get_ident
        except ImportError:
            from threading import get_ident as _get_ident

        call_key = (id(self), _get_ident())
        if call_key in _repr_running:
            # TODO?: make a YAML-like naming & referencing?
            # (too complicated for a repr() though)
            return '...'
        _repr_running[call_key] = 1
        try:
            return fn(self)
        finally:
            del _repr_running[call_key]


# https://pypi.python.org/pypi/ordereddict
# or /usr/lib/python2.7/collections.py
# with modifications
class OrderedDict(ODReprMixin, dict, DictMixin):

    __end = None
    __map = None

    def __init__(self, *args, **kwds):
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        # try:
        #     self.__end
        # except AttributeError:
        #     self.clear()
        if self.__end is None:
            # XX: What was that, an inheritance support?
            # (the original OrderedDict code even does an AttributeError catch)
            self.clear()
        # self.clear()
        self.update(*args, **kwds)

    def clear(self):
        self.__end = end = []
        # sentinel node for doubly linked list
        end += [None, end, end]
        # key --> [key, prev, next]
        self.__map = {}
        dict.clear(self)

    def __setitem__(self, key, value):
        if key not in self:
            end = self.__end
            curr = end[1]
            curr[2] = end[1] = self.__map[key] = [key, curr, end]
        dict.__setitem__(self, key, value)

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        key, prev, next_ = self.__map.pop(key)
        prev[2] = next_
        next_[1] = prev

    def __iter__(self):
        end = self.__end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.__end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def popitem(self, last=True):
        if not self:
            raise KeyError('dictionary is empty')
        if last:
            key = next(reversed(self))
        else:
            key = iter(self).next()
        value = self.pop(key)
        return key, value

    def __reduce__(self):
        items = [[k, self[k]] for k in self]
        tmp = self.__map, self.__end
        del self.__map, self.__end
        inst_dict = vars(self).copy()
        self.__map, self.__end = tmp
        if inst_dict:
            return (self.__class__, (items,), inst_dict)
        return self.__class__, (items,)

    def keys(self):
        return list(self)

    setdefault = DictMixin.setdefault
    update = DictMixin.update
    pop = DictMixin.pop
    values = DictMixin.values
    items = DictMixin.items

    # Will be provided in any version from whichever is available.
    iterkeys = getattr(DictMixin, 'iterkeys', None) or getattr(DictMixin, 'keys')
    itervalues = getattr(DictMixin, 'itervalues', None) or getattr(DictMixin, 'values')
    iteritems = getattr(DictMixin, 'iteritems', None) or getattr(DictMixin, 'items')

    def copy(self):
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def __eq__(self, other):
        if isinstance(other, OrderedDict):
            if len(self) != len(other):
                return False
            for p, q in zip(self.items(), other.items()):
                if p != q:
                    return False
            return True
        return dict.__eq__(self, other)

    def __ne__(self, other):
        return not self.__eq__(other)  # not self == other


# ###
# MultiValueDict from django.utils.datastructures
# ###

class MultiValueDictKeyError(KeyError):
    pass


class MultiValueDict(dict):
    """
    A subclass of dictionary customized to handle multiple values for the
    same key.

    >>> d = MultiValueDict({'name': ['Adrian', 'Simon'], 'position': ['Developer']})
    >>> d['name']
    'Simon'
    >>> d.getlist('name')
    ['Adrian', 'Simon']
    >>> d.getlist('doesnotexist')
    []
    >>> d.getlist('doesnotexist', ['Adrian', 'Simon'])
    ['Adrian', 'Simon']
    >>> d.get('lastname', 'nonexistent')
    'nonexistent'
    >>> d.setlist('lastname', ['Holovaty', 'Willison'])

    This class exists to solve the irritating problem raised by cgi.parse_qs,
    which returns a list for every key, even though most Web forms submit
    single name-value pairs.
    """
    def __init__(self, key_to_list_mapping=()):
        super(MultiValueDict, self).__init__(key_to_list_mapping)

    @classmethod
    def make_from_items(cls, items):
        key_to_list_mapping = {}
        for key, val in items:
            key_to_list_mapping.setdefault(key, []).append(val)
        return cls(key_to_list_mapping)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__,
                             super(MultiValueDict, self).__repr__())

    def __getitem__(self, key):
        """
        Returns the last data value for this key, or [] if it's an empty list;
        raises KeyError if not found.
        """
        try:
            list_ = super(MultiValueDict, self).__getitem__(key)
        except KeyError:
            raise MultiValueDictKeyError(repr(key))
        try:
            return list_[-1]
        except IndexError:
            # Does anyone know the reason for this behaviour and ways
            # there could happen to be an empty list?
            return []

    def __setitem__(self, key, value):
        super(MultiValueDict, self).__setitem__(key, [value])

    def __copy__(self):
        return self.__class__([
            (k, v[:])
            for k, v in self.lists()
        ])

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        result = self.__class__()
        memo[id(self)] = result
        for key, value in dict.items(self):
            dict.__setitem__(result, copy.deepcopy(key, memo),
                             copy.deepcopy(value, memo))
        return result

    def __getstate__(self):
        obj_dict = self.__dict__.copy()
        obj_dict['_data'] = {k: self.getlist(k) for k in self}
        return obj_dict

    def __setstate__(self, obj_dict):
        data = obj_dict.pop('_data', {})
        for k, v in data.items():
            self.setlist(k, v)
        self.__dict__.update(obj_dict)

    def get(self, key, default=None):
        """
        Returns the last data value for the passed key. If key doesn't exist
        or value is an empty list, then default is returned.
        """
        try:
            val = self[key]
        except KeyError:
            return default
        if val == []:
            return default
        return val

    def getlist(self, key, default=None):
        """
        Returns the list of values for the passed key. If key doesn't exist,
        then a default value is returned.
        """
        try:
            return super(MultiValueDict, self).__getitem__(key)
        except KeyError:
            if default is None:
                return []
            return default

    def setlist(self, key, list_):
        super(MultiValueDict, self).__setitem__(key, list_)

    def setdefault(self, key, default=None):
        if key not in self:
            self[key] = default
            # Do not return default here because __setitem__() may store
            # another value -- QueryDict.__setitem__() does. Look it up.
        return self[key]

    def setlistdefault(self, key, default_list=None):
        if key not in self:
            if default_list is None:
                default_list = []
            self.setlist(key, default_list)
            # Do not return default_list here because setlist() may store
            # another value -- QueryDict.setlist() does. Look it up.
        return self.getlist(key)

    def appendlist(self, key, value):
        """Appends an item to the internal list associated with key."""
        self.setlistdefault(key).append(value)

    def _iteritems(self):
        """
        Yields (key, value) pairs, where value is the last item in the list
        associated with the key.
        """
        for key in self:
            yield key, self[key]

    def _iterlists(self):
        """Yields (key, list) pairs."""
        return six.iteritems(super(MultiValueDict, self))

    def _itervalues(self):
        """Yield the last value on every key list."""
        for key in self:
            yield self[key]

    if six.PY3:
        items = _iteritems
        lists = _iterlists
        values = _itervalues
    else:
        iteritems = _iteritems
        iterlists = _iterlists
        itervalues = _itervalues

        def items(self):
            return list(self.iteritems())

        def lists(self):
            return list(self.iterlists())

        def values(self):
            return list(self.itervalues())

    def copy(self):
        """Returns a shallow copy of this object."""
        return copy.copy(self)

    def update(self, *args, **kwargs):
        """
        update() extends rather than replaces existing key lists.
        Also accepts keyword args.
        """
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))
        if args:
            other_dict = args[0]
            if isinstance(other_dict, MultiValueDict):
                for key, value_list in other_dict.lists():
                    self.setlistdefault(key).extend(value_list)
            else:
                try:
                    for key, value in other_dict.items():
                        self.setlistdefault(key).append(value)
                except TypeError:
                    raise ValueError("MultiValueDict.update() takes either a MultiValueDict or dictionary")
        for key, value in six.iteritems(kwargs):
            self.setlistdefault(key).append(value)

    def dict(self):
        """
        Returns current object as a dict with singular values.
        """
        return {key: self[key] for key in self}


# ######
# Other stuff
# ######


def _is_MultiValueDict(val, deep=False):
    """
    Check the object for MultiValueDict face (to support e.g. the
    django's one). Does not include MVODs.
    """
    if isinstance(val, MultiValueDict):
        return True
    # The interesting internal-data access method 'lists':
    if not hasattr(val, 'lists'):
        return False
    if deep:
        return any(
            cls.__name__ == 'MultiValueDict'
            for cls in val.__class__.__mro__)
    else:
        return isinstance(val, dict)


def _lists_group(items):
    """ items -> key_to_list_mapping """
    # Pretty much `pyaux.base.group()`
    result = {}
    for key, val in items:
        try:
            list_ = result[key]
        except KeyError:
            list_ = [val]
            result[key] = list_
        else:
            list_.append(val)
    return result


def _lists_group_ordered(items):
    """ items -> key_to_list_mapping keeping some order """
    order = []
    lists = {}
    for key, val in items:
        try:
            list_ = lists[key]
        except KeyError:
            lst = [val]
            lists[key] = lst
            order.append((key, lst))
        else:
            list_.append(val)
    return iter(order)


def _lists_ungroup(key_to_list_mapping):
    """" key_to_list_mapping -> items """
    if isinstance(key_to_list_mapping, dict):
        key_to_list_mapping = key_to_list_mapping.items()
    result = [
        (key, val)
        for key, vals in key_to_list_mapping
        for val in vals]
    return result


class MVOD_Common(ODReprMixin, object):

    _data_internal = ()

    # Conveniences

    _lists_group = staticmethod(_lists_group)
    _lists_group_ordered = staticmethod(_lists_group_ordered)
    _lists_ungroup = staticmethod(_lists_ungroup)

    def _preprocess_data(self, data):
        """
        Make sure the passed data is a list of pairs; returns a tuple
        with the pair-tuples.
        """
        res = []
        for i, item in enumerate(data):
            lv = len(item)  # Paranoidally avoid calling it twice.
            if lv != 2:
                raise ValueError((
                    "dictionary update sequence element #%d has"
                    " length %r; 2 is required") % (i, lv))
            key, val = item
            res.append((key, val))
        return tuple(res)

    def _process_upddata(self, args, kwds, preprocess=True, strict=False):
        """
        Convert some function call args into list of key-value pairs
        (same as `dict(*args, **kwds)` does).  Returns a tuple with
        pair-tuples.
        """
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        arg = args[0] if args else []
        if isinstance(arg, MVOD):  # support init / update from antother MVOD
            arg = arg._data
        elif _is_MultiValueDict(arg):  # Other `MultiValueDict`s
            arg = self._lists_ungroup(arg.lists())
        elif isinstance(arg, dict):
            arg = getattr(arg, 'iteritems', arg.items)()
        if kwds:
            if strict:
                raise TypeError('initializing an ordered dict from keywords is not recommended')
            # Append the keywords to the other stuff
            arg = itertools.chain(arg, kwds.items())
        if preprocess:
            arg = self._preprocess_data(arg)  ## NOTE: iterates over it and makes a tuple.
        return arg

    # A set of properties to set the data with or without checking it
    # for validity.

    @property
    def _data(self):
        return self._data_internal

    @_data.setter
    def _data(self, val):
        self._data_internal = self._preprocess_data(val)
        self._update_cache()

    @property
    def _data_checked(self):
        return self._data_internal

    @_data_checked.setter
    def _data_checked(self, val):
        assert isinstance(val, tuple)
        self._data_internal = val
        self._update_cache()

    def _update_cache(self):
        raise NotImplementedError

    # Commonly usable methods

    def clear(self):
        # does `self._update_cache`:
        self._data_checked = ()

    def __copy__(self):
        return self.__class__(self._data)

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        result = self.__class__()
        memo[id(self)] = result
        result._data_checked = copy.deepcopy(self._data, memo)
        return result

    copy = __copy__  # XXX: re-check.

    # TODO?: __getstate__, __setstate__; probably not useful.

    def _iteritems(self):
        # WARN: no dict-changed-while-iterating handling. In practice,
        # iteration is always over a copy (since self._data is a tuple).
        for item in self._data:
            yield item

    def __reversed__(self):
        # See the `_iteritems` dict-changed-while-iterating note.
        for key, _ in reversed(self._data):
            yield key

    def _iterlists(self, ordered=True):
        """ MultiValueDict-like (django) method. Not very optimal. """
        # See the `_iteritems` dict-changed-while-iterating note.
        pre_func = self._lists_group_ordered if ordered else self._lists_group
        for key, lst in pre_func(self._data):
            yield key, lst

    def _itervalues(self):
        for _, val in self._iteritems():
            yield val

    def _iterkeys(self):
        # Supports the duplicates.
        for key, _ in self._iteritems():
            yield key

    def __iter__(self):
        return self._iterkeys()

    # # six:

    if six.PY3:

        def items(self, *args, **kwargs):
            return self._iteritems(*args, **kwargs)

        def lists(self, *args, **kwargs):
            return self._iterlists(*args, **kwargs)

        def values(self, *args, **kwargs):
            return self._itervalues(*args, **kwargs)

        def keys(self, *args, **kwargs):
            return self._iterkeys(*args, **kwargs)

    else:

        def iteritems(self, *args, **kwargs):
            return self._iteritems(*args, **kwargs)

        def iterlists(self, *args, **kwargs):
            return self._iterlists(*args, **kwargs)

        def itervalues(self, *args, **kwargs):
            return self._itervalues(*args, **kwargs)

        def iterkeys(self, *args, **kwargs):
            return self._iterkeys(*args, **kwargs)

        def items(self, *args, **kwargs):
            return list(self.iteritems(*args, **kwargs))

        def lists(self, *args, **kwargs):
            return list(self.iterlists(*args, **kwargs))

        def values(self, *args, **kwargs):
            return list(self.itervalues(*args, **kwargs))

        def keys(self, *args, **kwargs):
            return list(self.iterkeys(*args, **kwargs))

    # Bit more copypaste from UserDict.DictMixin (because most other
    # methods from it are unnecessary)

    def pop(self, key, *args):
        if len(args) > 1:
            raise TypeError("pop expected at most 2 arguments, got %r" % (1 + len(args),))
        try:
            value = self[key]
        except KeyError:
            if args:
                return args[0]
            raise
        del self[key]
        return value

    def setdefault(self, key, failobj=None):
        if key not in self:
            self[key] = failobj
        return self[key]

    def popitem(self, last=True):
        if not self:
            raise KeyError('dictionary is empty')
        data = self._data
        if last:
            item = data[-1]
            self._data_checked = data[:-1]
        else:
            item = data[0]
            self._data_checked = data[1:]
        return item

    @classmethod
    def fromkeys(cls, iterable, value=None):
        return cls([(k, value) for k in iterable])

    def __eq__(self, other):
        """ ...

        WARN: `mvod == od` and `od == mvod` might have different
        results (because OD doesn't handle MVODs).
        """
        if isinstance(other, (MVOD_Common, OrderedDict)):
            # Quicker pre-check:
            if not dict.__eq__(self, other):
                return False
            # In the end it comes down to items.
            if self._data == tuple(other.items()):
                return True
            return False
        elif _is_MultiValueDict(other):
            return self._data == tuple(self._lists_ungroup(other.lists()))
        else:
            return dict.__eq__(self, other)

    def __ne__(self, other):
        return not self == other

    # __nonzero__ does not need overriding as dict handles that.

    # XXX: some other methods?


class MVOD(MVOD_Common, dict):
    """ MultiValuedOrderedDict: A not-very-optimized (most write operations
    are at least O(N) with the re-hashing cost) somewhat-trivial verison.
    Stores a tuple of pairs as the actual data (in `_data`), uses it for
    iteration, caches dict(data) as self for optimized key-access. """
    # TODO?: support unhashable keys (by skipping them in the cache)
    # TODO?: make the setitem behaviour configurable per instance

    def __init__(self, *args, **kwds):
        self.update(*args, **kwds)

    def _update_cache(self):
        dict.clear(self)
        dict.update(self, self._data)

    def update_append(self, *args, **kwds):
        """ ...
        (the usual equivalent of dict.update)

        WARNING: appends the data; thus, multiple `update`s will cause
        it to grow; thus, might need to be `.deduplicate`d.
        """
        data_new = self._process_upddata(args, kwds)
        # XXXXX: A dilemma:
        #  * If self._data is a list, then it can be mutated by the
        #    user without updating the cache
        #  * If self._data is a tuple, then it has to be re-created for
        #    any change (current version).
        #  * If self._data is a custom proxy to a list... dunno. TODO?
        #    (similar to the django.utils.datastructures.ImmutableList
        #    except not tuple-derived)
        self._data_checked = self._data + data_new

    def update_replace(self, *args, **kwds):
        """
        A closer equivalent of dict.update that removes all the
        previous occurrences of the keys that are updated and appends
        the new ones to the end. The new keys still can occur multiple
        times.
        """
        data_new = self._process_upddata(args, kwds)
        keys = set(key for key, val in data_new)
        pre_data = tuple((key, val) for key, val in self._data
                         if key not in keys)
        self._data_checked = pre_data + data_new

    def update_inplace(self, *args, **kwds):
        """ A closer equivalent of OrderedDict.update that replaces the
        previous occurrences at the point of first occurrence.  Does not yet
        support multiple items handling in the updated keys. """
        data_new = self._process_upddata(args, kwds)
        news = dict(*args, **kwds)
        # XXXX/TODO: To update with multiple values will have to do `MVOD(*args,
        #   **kwds).lists()`, and (configurably) pop either each item from the
        #   list when it occurs or just one item each time.
        pre_data = tuple((key, news.pop(key, val))
                         for key, val in self._data)
        # Add the non-previously-existing ones.
        data_new = [(key, val) for key, val in data_new if key in news]
        self._data_checked = pre_data + tuple(data_new)

    update = update_append

    def __delitem__(self, key):
        self._data_checked = tuple((k, v) for k, v in self._data if k != key)

    # def __getitem__:  inherited from `dict`

    def __setitem__(self, key, value):
        return self.update(((key, value),))
        # self._data_checked = self._data + ((key, value),)

    def deduplicate(self, how='last'):
        """ ...
        NOTE: this method is equivalent to deduplicate_last by default. """
        if how == 'first':
            pass
        elif how == 'last':
            return self.deduplicate_last()
        else:
            raise ValueError("Unknown deduplication `how`: %r" % (how,))
        self._data_checked = tuple(uniq_g(self._data, key=lambda item: item[0]))

    def deduplicate_last(self):
        data_pre = uniq_g(reversed(self._data), key=lambda item: item[0])
        data_pre = list(data_pre)
        self._data_checked = tuple(reversed(data_pre))

    def deduplicated(self, **kwa):
        cp = self.copy()
        cp.deduplicate(**kwa)
        return cp

    def getlist(self, key, default=None):
        values = [
            item_val for item_key, item_val in self._iteritems()
            if item_key == key]
        if values:
            return values
        if default is None:
            return []
        return default


class MVLOD(MVOD_Common, MultiValueDict):
    """
    A Multi-Value Ordered Dict with slow modification and fast
    value-list access.

    Implemented as MultiValueDict that stores the original/intended
    items list.

    NOTE: instantiated from `items`, not `key_to_list_mapping`. Use
    `make_from_key_to_list_mapping` for the MultiValueDict()'s
    behaviour.
    """

    _optimised = True

    def __init__(self, *args, **kwds):
        self.update(*args, **kwds)

    @classmethod
    def make_from_key_to_list_mapping(cls, key_to_list_mapping=()):
        return cls(cls._lists_ungroup(key_to_list_mapping))

    @classmethod
    def make_from_items(cls, items):
        return cls(items)

    # Core logic

    def _update_cache(self):
        dict.clear(self)
        dict.update(self, self._lists_group(self._data))

    def update_append(self, *args, **kwds):
        """ ...
        (the usual equivalent of dict.update)

        WARNING: appends the data; thus, multiple `update`s will cause
        it to grow; thus, might need to be `.deduplicate`d.
        """
        data_new = self._process_upddata(args, kwds)
        data_result = self._data + data_new
        if not self._optimised:
            self._data_checked = data_result
        else:
            self._data_internal = data_result
            ktlm_new = self._lists_group(data_new)
            for key, list_ in ktlm_new.items():
                try:
                    target_list = dict.__getitem__(self, key)
                except KeyError:
                    dict.__setitem__(self, key, list_)
                else:
                    target_list.extend(list_)

    def update_replace(self, *args, **kwds):
        """
        A closer equivalent of dict.update that removes all the
        previous occurrences of the keys that are updated and appends
        the new ones to the end. The new keys still can occur multiple
        times.
        """
        data_new = self._process_upddata(args, kwds)
        keys = set(key for key, val in data_new)
        data_kept = tuple(
            (key, val) for key, val in self._data
            if key not in keys)
        data_result = data_kept + data_new
        if not self._optimised:
            self._data_checked = data_result
        else:  # Optimised cache-dict mangle
            self._data_internal = data_result
            ktlm_new = self._lists_group(data_new)
            # # Should be unnecessary:
            # for key in keys:
            #     self.pop(key, None)
            for key, list_ in ktlm_new.items():
                dict.__setitem__(self, key, list_)
                # super(MVLOD, self).setlist(key, list_)

    def setlist(self, key, list_):
        data_new = self._lists_ungroup([(key, list_)])
        self.update_replace(data_new)

    def appendlist(self, key, value):
        self.update_append([(key, value)])

    def __delitem__(self, key):
        data_result = tuple(
            (item_key, item_val) for item_key, item_val in self._data
            if item_key != key)
        # See if we removed nothing.
        if len(data_result) == len(self._data):
            raise KeyError(key)
        if not self._optimised:
            self._data_checked = data_result
        else:
            self._data_internal = data_result
            try:
                dict.__delitem__(self, key)
            except KeyError:
                raise Exception("The key should've been there", self, key)

    # The defaults (for setitem, instantiation)

    update = update_append

    def __setitem__(self, key, value):
        self.update([(key, value)])

    # Other optimisations

    def _iterlists(self, ordered=False):
        if ordered:
            return super(MVLOD, self)._iterlists(ordered=ordered)
        if six.PY3:
            return dict.items(self)
        else:
            return dict.iteritems(self)


# NOTE: A simpler form is available in the `base`
class dotdictx(dict):
    """ A dict subclass with items also available over
    attributes. Skips the attributes starting with '_' (for
    mixinability) """

    def __getattr__(self, name):
        if name.startswith('__'):  # NOTE: two underscores.
            return super(dotdictx, self).__getattr__(name)
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return super(dotdictx, self).__setattr__(name, value)
        self[name] = value


class DotOrderedDict(dotdictx, OrderedDict):
    """ ... """


class defaultdictx(dict):
    """ A purepython & simplified `defaultdict` version that doesn't change
    the interface except for the `_default` attribute (with the
    default_factory) """
    _default = None

    def __getitem__(self, name):
        try:
            return super(defaultdictx, self).__getitem__(name)
        except KeyError:
            if self._default is not None:
                val = self._default()
                self[name] = val
                return val
            raise


def hasattr_x(obj, name):
    """ A safer `hasattr` that only checks for simple attributes
    rather than properties or __getattr__ results. """
    try:
        object.__getattribute__(obj, name)
    except AttributeError:
        return False
    return True


class DefaultDotDictMixin(dotdict, defaultdictx):
    """ A class that tries to combine DefaultDict and dotdict without causing
    too much of a mess. NOTE: skips _attributes on setattr and __attributes on
    getattr. """

    def __getattr__(self, name):
        # Mostly necessary to avoid `defaultdict`ing some special
        # methods like __getstate__ that weren't defined on the class.
        # (could disable the defaultdict'ing for that, though)
        if name.startswith('__'):  # NOTE: two underscores.
            return self.__getattribute__(name)  # Basically `raise AttributeError`.
        return super(DefaultDotDictMixin, self).__getattr__(name)  # __getitem__

    def __setattr__(self, name, value):
        # if name in self.__dict__:
        # if hasattr_x(self, name):
        #     self.__dict__[name] = value
        #     return

        # Mostly necessary for class code that sets attributes like
        # '_OrderedDict__end' (or the defaultdictx._default)
        if name.startswith('_'):
            # Basically `object.__setattr__(…)`
            # WARN: querying `d._attr` still sets it to the
            # default. Not necessarily problematic though.
            return dict.__setattr__(self, name, value)

        return super(DefaultDotDictMixin, self).__setattr__(name, value)  # __setitem__


class dodd(DefaultDotDictMixin, OrderedDict):
    """ DotOrderedDefaultDict. Set `_default` attribute on it to a factory to
    use it as a defaultdict. NOTE: ignores attributes starting with '_' """
    pass


class mvdodd(DefaultDotDictMixin, MVOD):
    pass


class redodd(dodd):
    """ Recursive dodd (by default) """
redodd._default = redodd


class remvdodd(mvdodd):
    """ ... """
remvdodd._default = remvdodd


# TODO: test... things...

# TODO?: Add some other sort of multivaluedness
# (django.utils.datastructures.MultiValueDict).  (and maybe add a
# separate class compatible with id and/or the django.http.QueryDict
# or something)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# pylint: disable=broad-except
""" yaml -> json for pretty-writing. """

# import os
import sys
import argparse
try:
    import simplejson as json
except Exception:
    import json
import yaml
from pyaux.base import to_str


def cmd_make_parser():
    parser = argparse.ArgumentParser(
        description=(
            "yaml -> json for pretty-writing"
        )
    )
    parser.add_argument(
        '-nu',
        dest='ensure_ascii', action='store_true',
        default=False,
        help="Do not allow unicode in output (reverse of json's 'ensure_ascii')",
    )
    parser.add_argument(
        '-i', '--indent',
        dest='indent',
        default=None,
        help="Indentation ('-' to disable pretty-formatting)",
    )
    parser.add_argument(
        '-c', '--compact',
        dest='compact', action='store_true',
        default=False,
        help="Compact representation (without some of the extra whitespaces)",
    )

    return parser


def main():

    parser = cmd_make_parser()
    params = parser.parse_args()

    # TODO?: support input file
    data_in = sys.stdin.read()

    def bailout(msg):
        sys.stderr.write(
            "ERROR: fjson_yaml: %s; original data as follows (on stdout)\n" % (msg,))
        sys.stdout.write(data_in)
        return 13

    Loader = getattr(yaml, 'CSafeLoader', None) or yaml.SafeLoader
    try:
        data_data = yaml.load(data_in, Loader=Loader)
    except Exception as exc:
        return bailout("Error parsing as yaml: %s" % (exc,))

    json_kwargs = dict(
        ensure_ascii=params.ensure_ascii,
    )
    if params.indent is None:
        json_kwargs['indent'] = None if params.compact else 2
    elif params.indent == '-':
        json_kwargs['indent'] = None
    else:
        json_kwargs['indent'] = int(params.indent)
    if params.compact:
        json_kwargs['separators'] = (',', ':')

    try:
        out = json.dumps(data_data, **json_kwargs)
    except Exception as exc:
        return bailout("Error dumping as json: %s" % (exc,))

    # TODO?: support output file
    sys.stdout.write(to_str(out))
    if out[-1] != '\n':  # Just in case
        sys.stdout.write('\n')
    sys.stdout.flush()


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
''' Serialization formats converter '''

from __future__ import print_function, unicode_literals, absolute_import, division

import sys
import argparse
import functools
from pyaux.iterables import prefetch_first


def _bytes_stdin():
    return getattr(sys.stdin, 'buffer', sys.stdin)


def _bytes_stdout():
    return getattr(sys.stdout, 'buffer', sys.stdout)


# ARG_TRUES = ('yes', 'true', 't', 'y', 1)
# ARG_FALSES = ('no', 'false', 'f', 'n', '0')
# ARG_NULLS = ('', 'null', 'none')
# def argument_bool(value):
#     value = value.lower()
#     if value in ARG_TRUES:
#         return True
#     if value in ARG_FALSES:
#         return False
#     raise argparse.ArgumentTypeError('Boolean value expected.')
# def argument_ternary(value):
#     value = value.lower()
#     if value in ARG_TRUES:
#         return True
#     if value in ARG_FALSES:
#         return False
#     if value in ARG_NULLS:
#         return None
#     raise argparse.ArgumentTypeError('Boolean value expected.')


def cmd_make_parser(**kwa):
    parser = argparse.ArgumentParser(
        description='Serialization formats converter')
    parser.add_argument(
        '-i', '--input',
        default=_bytes_stdin(),
        type=argparse.FileType('rb'),
        help='Source file; empty to use stdin',
    )
    parser.add_argument(
        '-f', '--input-format',
        default='auto',
        choices=('auto', 'json', 'msgp', 'msgpack', 'yaml'),
        help='Serialization format of the input',
    )
    parser.add_argument(
        '-cp', '--input-encoding',
        default='utf-8',
        help=(
            'Character encoding for the input;'
            ' use "" to disable when possible;'
            ' mostly for msgpack'
        ),
    )
    parser.add_argument(
        '-t', '--output-format',
        default='yaml',
        choices=('json', 'msgp', 'msgpack', 'yaml'),
        help='Serialization format for the input',
    )
    parser.add_argument(
        '-ocp', '--output-encoding',
        default='utf-8',
        help='Character encoding for the output',
    )
    parser.add_argument(
        '-o', '--output',
        default=_bytes_stdout(),
        type=argparse.FileType('wb'),
        help='Destination file; empty to use stdout',
    )

    parser.add_argument(
        '-od', '--original-data-on-failure',
        dest='original_data_on_failure', action='store_true',
        help=(
            "Dump the input data on (de)serialization failure"
            " (useful when parsing the output with the eyes)"
            " (default)"
        ),
    )
    parser.add_argument(
        '-nod', '--no-original-data-on-failure',
        dest='original_data_on_failure', action='store_false',
        help="Do not dump the input data on (de)serialization failure")
    parser.set_defaults(original_data_on_failure=True)

    # # YAML
    parser.add_argument(
        '-yf', '--no-default-flow-style',
        dest='default_flow_style', action='store_false',
        help="pyyaml's `default_flow_style=False` (generally larger output) (default is None)",
    )
    parser.add_argument(
        '-yx', '--default-flow-style',
        dest='default_flow_style', action='store_true',
        help="pyyaml's `default_flow_style=True` (more json-like) (default is None)",
    )
    parser.set_defaults(default_flow_style=None)

    # # YAML/JSON
    parser.add_argument(
        '-c', '--color',
        choices=('yes', 'always', 'no', 'never', 'auto'),
        default='auto',
        help=(
            "Colorize the output (requires pygments) (default: 'auto')"
            " (`ls --color` semantics)"
        ),
    )
    parser.add_argument(
        '-u', '--allow-unicode', dest='allow_unicode',
        default='yes', choices=('yes', 'no'),
        help=(
            "Allow non-ascii in output"
            " (yaml's 'allow_unicode', json's 'ensure_ascii')"
        ),
    )
    # # TODO:
    # parser.add_argument(
    #     '-nu',
    #     dest='allow_unicode', action='store_false',
    #     help="Do not allow non-ascii in output (shorthand)",
    # )
    parser.set_defaults(allow_unicode=True)

    parser.add_argument(
        '--indent',
        type=int, default=2,
        help=(
            "Formatting indent; use 0 for compact JSON"
            " (Note: pyyaml seems to ignore indent outside [2, 9])"
        ),
    )

    # Currently JSON-only.
    parser.add_argument(
        '--sk', '--sort-keys', dest='sort_keys',
        default='auto', choices=('yes', 'no', 'auto'),
        help="Sort the keys in mappings (default: yes)")

    return parser


class Bailout(Exception):
    """ Cannot do as requested """


def main():
    parser = cmd_make_parser()
    params = parser.parse_args()
    # TODO: streamed/chunked/prefetched/whatever.
    data_in = params.input.read()
    if not data_in:
        sys.stderr.write('Empty input.\n')
        return 17
    try:
        return main_i(data_in, params)
    except Bailout as exc:
        msg = exc.args[0]
        err_msg_pieces = ['ERROR: f_convert: {}'.format(msg)]
        if params.original_data_on_failure:
            err_msg_pieces.append('; original data as follows (on stdout):')
        err_msg_pieces.append('\n')
        sys.stderr.write(''.join(err_msg_pieces))
        if params.original_data_on_failure:
            _bytes_stdout().write(data_in)
        return 13


def get_json():
    """ Only two versions support most of the required parameters """
    try:
        import simplejson as json
    except Exception:
        import json
    return json


def parse_json(data_in):
    json = get_json()
    # TODO: support json-lines.
    # (but might have to switch to streaming completely for that)
    yield json.loads(data_in)


def parse_yaml(data_in):
    import yaml
    Loader = getattr(yaml, 'CSafeLoader', None) or yaml.SafeLoader
    # TODO: support multi-document streams
    return yaml.load_all(data_in, Loader=Loader)


def parse_msgp(data_in, input_encoding='utf-8'):
    import msgpack
    kwargs = {}
    if input_encoding:
        kwargs.update(encoding=input_encoding)

    # TODO: proper streaming
    try:
        from io import BytesIO as iowrap
    except Exception:
        from cStringIO import StringIO as iowrap
    data_in = iowrap(data_in)

    stream = msgpack.Unpacker(data_in, **kwargs)
    try:
        item = next(stream)
    except StopIteration:
        return
    if isinstance(item, (int, float)):
        raise Exception("msgpack returned a number which is suspicious", item)
    yield item
    for item in stream:
        yield item


def prefetch_first_wrap(func, count=1, require=True):

    @functools.wraps(func)
    def prefetch_first_wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        return prefetch_first(result, count=count, require=require)

    return prefetch_first_wrapped


def parse_auto(data_in):
    errors = {}
    # TODO: streaming support which would require pre-peeking into data_in.
    try:
        return prefetch_first(parse_json(data_in), require=True)
    except Exception as exc:
        errors.update(json_error=exc)
    try:
        return prefetch_first(parse_yaml(data_in), require=True)
    except Exception as exc:
        errors.update(yaml_error=exc)
    try:
        return prefetch_first(parse_msgp(data_in), require=True)
    except Exception as exc:
        errors.update(msgp_error=exc)
    try:
        return prefetch_first(parse_msgp(data_in, input_encoding=None))
    except Exception as exc:
        errors.update(msgp_bytes_error=exc)
    raise Exception(errors)


def dump_jsons(items, colorize=False, **kwargs):
    json = get_json()
    for item in items:
        res = json.dumps(item, **kwargs)
        # TODO: colorize
        yield res


def dump_yamls(items, colorize=False, **kwargs):
    import yaml
    if colorize:
        from pyaux.base import colorize_yaml
    Dumper = getattr(yaml, 'CSafeDumper', None) or yaml.SafeDumper
    # TODO: streamed
    result = yaml.dump_all(items, Dumper=Dumper, **kwargs)
    if colorize:
        if isinstance(result, bytes):
            result = result.decode('utf-8')
        result = colorize_yaml(result)
    yield result


def dump_msgp(items, use_bin_type=True, **kwargs):
    import msgpack
    for item in items:
        yield msgpack.packb(
            item,
            use_bin_type=use_bin_type,
            **kwargs)


def bailout_wrap(func, err_tpl):

    @functools.wraps(func)
    def bailout_wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            raise Bailout(err_tpl.format(exc=exc))

    return bailout_wrapped


def make_parse_func(params):
    input_format = params.input_format

    if input_format == 'auto':
        parse_func = parse_auto
        err_tpl = 'Error parsing as JSON/YAML/MSGPACK: {exc}'
    elif input_format == 'json':
        parse_func = parse_json
        err_tpl = 'Error parsing as JSON: {exc}'
    elif input_format == 'yaml':
        parse_func = parse_yaml
        err_tpl = 'Error parsing as YAML: {exc}'
    elif input_format in ('msgp', 'msgpack'):
        parse_func = functools.partial(
            parse_msgp,
            input_encoding=params.input_encoding)
        err_tpl = 'Error parsing as MsgPack: {exc}'
    else:
        raise Exception("Unknown input_format")

    parse_func = prefetch_first_wrap(parse_func)
    parse_func = bailout_wrap(parse_func, err_tpl)

    return parse_func


def make_outs_func(params):
    isatty = params.output is _bytes_stdout() and sys.stdout.isatty()

    output_format = params.output_format
    colorize = (
        params.color == 'yes' or
        (params.color == 'auto' and isatty))
    # TODO: add the 'auto => unsorted in py3.7+' feature for JSON and YAML
    sort_keys = params.sort_keys in ('yes', 'auto')

    if output_format == 'json':
        outs_func = functools.partial(
            dump_jsons,
            ensure_ascii=not params.allow_unicode,
            colorize=colorize,
            sort_keys=sort_keys,
            indent=params.indent or None,
            # separators=(',',':'),
            # default=repr,
        )
    elif output_format == 'yaml':

        # TODO: overridable 'width' parameter.
        # Try using the whole terminal width:
        width = None
        try:
            import shutil
            # magic '8'
            width = shutil.get_terminal_size().columns - 8
        except Exception:  # pylint: disable=broad-except
            pass

        outs_func = functools.partial(
            dump_yamls,
            allow_unicode=params.allow_unicode,
            colorize=colorize,
            indent=params.indent or 2,
            default_flow_style=params.default_flow_style,
            width=width,
        )
    elif output_format in ('msgp', 'msgpack'):
        outs_func = dump_msgp
    else:
        raise Exception("Unknown output_format")

    err_tpl = 'Error serializing as {}: {{exc}}'.format(output_format)
    outs_func = prefetch_first_wrap(outs_func)
    outs_func = bailout_wrap(outs_func, err_tpl)

    return outs_func


def main_i(data_in, params):

    parse_func = make_parse_func(params)
    outs_func = make_outs_func(params)

    datas = parse_func(data_in)
    outs_data = outs_func(datas)

    output = params.output

    out_item = None
    for out_item in outs_data:
        if not isinstance(out_item, bytes):
            out_item = out_item.encode(params.output_encoding)
        output.write(out_item)

    if out_item is not None and out_item[-1] != b'\n':
        output.write(b'\n')

    return 0


if __name__ == '__main__':
    sys.exit(main())



# coding: utf8
"""
Additional utils for working with subprocesses.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import datetime
import subprocess

from .base import monotonic_now


def _out_cb_default_common(tag, line, timestamp=None, encoding='utf-8', errors='replace'):
    # timestamp_dt = datetime.datetime.fromtimestamp(timestamp)
    # # Less precise, more convenient.
    timestamp_dt = datetime.datetime.now()
    line = line.decode(encoding, errors=errors)
    if line.endswith('\n'):
        line = line[:-1]
    else:  # disambiguate
        line += '\\'
    print(
        timestamp_dt.strftime('%H:%M:%S'),
        tag,
        line)


def stdout_cb_default(line, timestamp=None, **kwargs):
    return _out_cb_default_common('O:', line, timestamp=timestamp)


def stderr_cb_default(line, timestamp=None, **kwargs):
    return _out_cb_default_common('E:', line, timestamp=timestamp)


class ProcessTimeoutError(Exception):
    """ ... """


class NonzeroExit(Exception):
    """ ... """


def set_fd_nonblocking(fdesc):
    import fcntl
    from os import O_NONBLOCK
    flags = fcntl.fcntl(fdesc, fcntl.F_GETFL)
    fcntl.fcntl(fdesc, fcntl.F_SETFL, flags | O_NONBLOCK)
    return fdesc


def poll_fds(
        fdmapping,
        log=lambda msg, *args: None,
        nonblocking=False,
        inner_timeout=1.0,
        total_timeout=None):
    """
    Given a `name -> fd` mapping, polls all the FDs until they are done
    and callbacks with (name, timestamp, data) tuples.

    If `nonblocking` is enabled it will try to read the data as it
    comes; otherwise it will read the data in lines (which can be less
    precise but more useful).

    Note that for precise timestamps in the result you need to minimize the
    output processing time.
    """
    import select

    start_time = monotonic_now()
    end_time = None
    if total_timeout:
        end_time = start_time + total_timeout

    # NOTE: also used to keep the current ones (removes fds as they
    # are done).
    fdrev = {fd: name for name, fd in fdmapping.items()}

    if nonblocking:
        for fdesc in fdrev:
            set_fd_nonblocking(fdesc)

    while True:
        if not fdrev:
            break
        fdlist = list(fdrev)
        timeout = inner_timeout
        if end_time:
            timeout = min(end_time - monotonic_now(), timeout)
            if timeout < 0.00001:
                raise ProcessTimeoutError(total_timeout)

        ready_fds, _, _ = select.select(fdlist, (), (), timeout)
        for fdesc in ready_fds:
            tag = fdrev[fdesc]
            timestamp = monotonic_now()
            if nonblocking:
                try:
                    data = fdesc.read()
                except OSError:
                    data = None
            else:
                data = fdesc.readline()
            if not data:
                log("fd is empty: %r / %r", tag, fdesc)
                fdrev.pop(fdesc)
                continue
            log("output (%r, %r, %r)", tag, timestamp, data)
            yield tag, timestamp, data


def run_cmd(*args, **kwargs):

    timeout = kwargs.pop('timeout', 600)
    stdout_cb = kwargs.pop('stdout_cb', stdout_cb_default)
    stderr_cb = kwargs.pop('stderr_cb', stderr_cb_default)
    nonblocking = kwargs.pop('nonblocking', False)

    tag_to_cb = dict(stdout=stdout_cb, stderr=stderr_cb)

    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)

    tag_to_fd = dict(stdout=proc.stdout, stderr=proc.stderr)
    outputs = poll_fds(tag_to_fd, nonblocking=nonblocking, total_timeout=timeout)
    for tag, timestamp, data in outputs:
        tag_to_cb[tag](data, timestamp=timestamp, proc=proc, outputs=outputs)

    ret = proc.poll()
    # Might or might not be required for the pipes to get closed:
    proc.communicate()
    if ret:
        raise NonzeroExit(ret)

    # ...


def main():
    """ A simple form of 'annotate-output' """
    # cmd = ('ping', '-c', '4', '-i', '0.2', '1.1.1.1')
    # cmd = ('sh', '-c', 'date -Ins; printf "zxcv"; sleep 5; printf "qwer\n"; date -Ins')
    import sys
    cmd = sys.argv[1:]
    run_cmd(*cmd, nonblocking=True)


if __name__ == '__main__':
    # python -m pyaux.procs ping -c 4 -i 0.2 1.1.1.1
    main()



# coding: utf8
""" Various modules for use in older python versions which don't have them natively yet """

import re

# Py3.3's shlex.quote piece backport

_sh_find_unsafe = re.compile(r'[^\w@%+=:,./-]').search


def sh_quote(s):
    if not s:
        return "''"
    if _sh_find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


def sh_quote_prettier(s):
    r"""
    Quote a value for copypasteability in a posix commandline.

    A more readable version than the backported one.

    >>> sh_quote_prettier("'one's one'")
    "\\''one'\\''s one'\\'"
    """
    if not s:
        return "''"
    if _sh_find_unsafe(s) is None:
        return s

    # A shorter version: backslash-escaped single quote.
    result = "'" + s.replace("'", "'\\''") + "'"
    # Cleanup the empty excesses at the ends
    _overedge = "''"
    if result.startswith(_overedge):
        result = result[len(_overedge):]
    if result.endswith(_overedge):
        result = result[:-len(_overedge)]
    return result



#!/usr/bin/env python
""" json -> yaml for pretty-reading. """

import os
import sys
import json
import yaml
import argparse
from pyaux.base import colorize_yaml as colorize
from pyaux.base import to_str


def cmd_make_parser(**kwa):
    parser = argparse.ArgumentParser(
        description=(
            "json -> yaml for pretty-reading"
        )
    )
    parser.add_argument(
        '-f', '--no-default-flow-style',
        dest='default_flow_style', action='store_false',
        default=None,
        help="pyyaml's `default_flow_style=False` (generally longer)",
    )
    parser.add_argument(
        '-x', '--default-flow-style',
        dest='default_flow_style', action='store_true',
        default=None,
        help="pyyaml's `default_flow_style=True` (more json-like)",
    )
    parser.add_argument(
        '-c', '--color',
        choices=('yes', 'always', 'no', 'never', 'auto'),
        default='auto',
        nargs='?',
        help=("Colorize the output (requires pygments) (default: 'auto')"
              " (`ls --color` semantics)"),
    )
    # parser.add_argument(
    #     '-c', dest='color', action='store_true',
    #     help="Colorize the output (requires pygments) (same as `--color=yes`)",
    # )
    parser.add_argument(
        '-nu',
        dest='allow_unicode', action='store_false',
        default=True,
        help="Do not allow unicode in output (pyyaml's 'allow_unicode')",
    )

    return parser


def main():

    parser = cmd_make_parser()
    params = parser.parse_args()

    # TODO?: support input file
    data_in = sys.stdin.read()

    def bailout(msg):
        sys.stderr.write(
            "ERROR: fjson_yaml: %s; original data as follows (on stdout)\n" % (msg,))
        sys.stdout.write(data_in)
        return 13

    try:
        data_data = json.loads(data_in)
    except Exception as exc:
        return bailout("Error parsing as json: %s" % (exc,))

    kwa = dict(
        default_flow_style=params.default_flow_style,
        allow_unicode=params.allow_unicode,
    )
    try:
        kwa['width'] = int(os.environ['WIDTH'])
    except Exception:
        pass

    try:
        out = yaml.safe_dump(data_data, **kwa)
    except Exception as exc:
        return bailout("Error dumping as yaml: %s" % (exc,))

    # Result:
    #   `--color=no` always skips this
    #   `--color`, `--color=yes`, `-c` always do this
    #   ``, `--color=auto` make this isatty-conditional.
    if ((params.color == 'auto' and sys.stdout.isatty()) or
            # NOTE: `'auto'` is default, `None` means it was
            # specified without an argument (equals to 'always')
            params.color in ('yes', 'always', True, None)):

        # pygments doesn't like utf-8 as-is; make it unicode
        if isinstance(out, bytes):
            out = out.decode('utf-8')

        # TODO?: support --color=auto
        out = colorize(out)

    # TODO?: support output file
    sys.stdout.write(to_str(out))
    if out[-1] != '\n':  # Just in case
        sys.stdout.write('\n')
    sys.stdout.flush()


if __name__ == '__main__':
    sys.exit(main())



# coding: utf8
""" Various functions for easier working with URLs """

from __future__ import absolute_import
from six.moves import urllib_parse as urlparse
from six import text_type as unicode
from .base import to_bytes, mangle_dict

urlencode = urlparse.urlencode


__all__ = (
    'url_to_querydict',
    'url_replace',
    'mangle_url',
    'mangle_url_m',
    'mangle_url_l',
)


def dict_to_bytes(d):
    """ Encode the strings in keys and values of a dict 'd' into
    bytestrings; useful for e.g. urlencoding the result """
    return {to_bytes(k): to_bytes(v) for k, v in d.items()}


def url_to_querydict(url):
    """ A shorthand for getting an url's query as a dict """
    url = to_bytes(url)
    return dict(urlparse.parse_qsl(urlparse.urlparse(url).query))


def url_replace(url, **params):
    """ Replace some named parts in an url; See `urlparse.ParseResult` for the names """
    url_fields = urlparse.ParseResult._fields
    name_to_num = {field: idx for idx, field in enumerate(url_fields)}
    url_parts = list(urlparse.urlparse(url))  # Need a copy anyway
    for key, val in params.items():

        # Allow supplying various stuff as a query
        if key == 'query' and not isinstance(val, (bytes, unicode)):
            if isinstance(val, dict):
                val = val.items()
            val = [(to_bytes(query_key), to_bytes(query_val))
                   for query_key, query_val in val]
            val = urlencode(val)

        num = name_to_num[key]  # Will except here if supplied an unknown url param
        url_parts[num] = val
    return urlparse.urlunparse(url_parts)


def mangle_url(url, include=None, exclude=None, add=None):
    """ 'mangle_dict' for url's query parameters """
    query = url_to_querydict(url)
    query_new = mangle_dict(query, include=include, exclude=exclude, add=add)
    return url_replace(url, query=query_new)


def mangle_url_m(url, include=None, exclude=None, add=None):
    """ Multivalue-version of the `mangle_url`; works with dict of lists only
    (`param -> [value1, …]`); sorts the resulting query """
    url = to_bytes(url)
    url_parts = urlparse.urlparse(url)
    ## NOTE: the order of the fields is still lost.
    query = urlparse.parse_qs(url_parts.query, keep_blank_values=1)
    query_new = mangle_dict(query, include=include, exclude=exclude, add=add)
    query_new = [(k, v) for k, vl in query_new.items() for v in vl]
    query_new = [(to_bytes(k), to_bytes(v)) for k, v in query_new]
    query_new = sorted(query_new)  # make the order stable since it's lost anyway
    return url_replace(url, query=query_new)


def mangle_url_l(url, include=None, exclude=None, add=None, **kwargs):
    """ mangle_url with preserving as much as possible (order, multiple values, empty values).

    Additional keyword parameters are passed to url_replace.

    >>> from pyaux.dicts import MVOD
    >>> query = r'a=&a=1&a=true&b=null&b=undefined&b=&b=5'
    >>> urlencode(MVOD(urlparse.parse_qsl(query, keep_blank_values=1))) == query
    True
    """
    from pyaux.dicts import MVOD
    url_parts = urlparse.urlparse(to_bytes(url))
    query = MVOD(urlparse.parse_qsl(url_parts.query, keep_blank_values=1))
    query_new = mangle_dict(
        query,
        include=include, exclude=exclude, add=add,
        _return_list=True)
    return url_replace(url, query=query_new, **kwargs)



#!/usr/bin/env python

import os
import sys
import json

from pyaux.base import to_bytes


def main():
    import msgpack
    indent = int(os.environ.get('INDENT') or '2')
    try:
        data_in = sys.stdin.buffer
    except AttributeError:
        data_in = sys.stdin

    try:
        stream = msgpack.Unpacker(data_in)  # , encoding="utf-8")
        for item in stream:
            try:
                data_out = json.dumps(item, indent=indent, sort_keys=True, ensure_ascii=False)
            except Exception as exc:
                data_out = "# (json failed: %r)  # %r" % (exc, item)
            sys.stdout.write(data_out)
            sys.stdout.write("\n")
            sys.stdout.flush()
    except Exception as exc:
        sys.stderr.write("#  -!!---- %s\n" % (exc,))
        sys.stdout.write(repr(data_in.read()))
        return 13


if __name__ == '__main__':
    sys.exit(main())



# coding: utf8
"""
Helpers for using Highcharts / Highstock in an IPython notebook.
"""

import copy
import json
import random
import string
import time
import datetime
from IPython.display import HTML
from pyaux.base import dict_merge, group


def mk_uid():
    return ''.join(
        random.choice(string.ascii_uppercase + string.digits)
        for _ in range(15))


def Highcharts_old(
        chart_def=None, chart_def_json=None, height=400, min_width=400, uid=None,
        highstock=True):
    assert chart_def or chart_def_json
    unique_id = mk_uid() if uid is None else uid
    chart_def_json = json.dumps(chart_def) if chart_def_json is None else chart_def_json
    if highstock:
        hsscript = "http://code.highcharts.com/stock/highstock.js"
    else:
        hsscript = "http://code.highcharts.com/highcharts.js"
    context = dict(
        chart_def_json=chart_def_json, chart_def=chart_def,
        min_width=min_width, height=height,
        unique_id=unique_id, hsscript=hsscript,
        hstag="'StockChart', " if highstock else "",
    )
    html = '''
    <script src="%(hsscript)s"></script>
    <script src="http://code.highcharts.com/modules/exporting.src.js"></script>

    <div id="chart_%(unique_id)s" style="min-width: %(min_width)spx; height: %(height)ipx; margin: 0 auto">Re-run cell if chart is not shown ...</div>
    <script>
        do_chart_%(unique_id)s = function() {
            $('#chart_%(unique_id)s').highcharts(%(hstag)s%(chart_def_json)s);
        }
        setTimeout("do_chart_%(unique_id)s()", 50)
    </script>
    ''' % context
    res = HTML(html)
    res.context = context
    return res




def RunJS(js, delayed=50):
    context = dict(js=js, unique_id=mk_uid(), delayed=delayed)
    if not delayed:
        html = '''
        <script>
        %(js)s
        </script>
        ''' % context
    else:
        html = '''
        <script>
            tmp_run_%(unique_id)s = function() {
                %(js)s
            }
            setTimeout("tmp_run_%(unique_id)s()", %(delayed)d)
        </script>
        ''' % context
    return HTML(html)


def Highcharts(
        chart_def=None,
        width=1800,
        height=800,
        highstock=True,
        **kwargs):
    if highstock:
        from highcharts import Highstock
        chart = Highstock()
    else:
        from highcharts import Highchart
        chart = Highchart()
    chart_def = chart_def.copy()
    chart_def['chart'] = (chart_def.get('chart') or {}).copy()
    chart_def['chart']['width'] = width
    chart_def['chart']['height'] = height

    series = chart_def.pop('series')
    chart.set_dict_options(chart_def)

    for line in series:
        chart.add_data_set(**line)

    # Note: this one essentially uses an iframe (in jupyter).
    return chart


def mk_chart_def(
        df=None,
        kwa=None,
        series=None,
        chart_type='line',
        timestamp_in_idx='auto',
        margin_right=130,
        margin_bottom=25,
        title='',
        subtitle='',
        xlabel='',
        ylabel='',
        zip_idx=True):
    """
    Convert a Pandas dataframe (or something else) to a highcharts
    chart definition.

    Sample serie: `dict(name='serie_name', data=[1.1, 2.2, 3.3])`.
    """
    series = [] if series is None else copy.copy(series)
    res = dict(
        chart={},
        title={},
        subtitle={},
        xAxis={},
        yAxis={},
        tooltip={},
        legend=dict(
            layout='vertical',
            align='right',
            verticalAlign='top',
            x=-10,
            y=100,
            borderWidth=0,
        ),
        plotOptions=dict(
            series=dict(
                animation=False,
                marker=dict(
                    enabled=True,
                ),
            ),
        ),
        series=series,
    )
    if chart_type is not None:
        res['chart']['type'] = chart_type
    if margin_right is not None:
        res['chart']['marginRight'] = margin_right
    if margin_bottom is not None:
        res['chart']['marginBottom'] = margin_bottom
    if title:
        res['title'].update(
            text=title,
            # center, supposedly
            x=-20)
    if subtitle:
        res['subtitle'].update(
            text=subtitle,
            x=-20)
    if xlabel:
        res['xAxis']['title'] = dict(text=xlabel)
    if ylabel:
        res['yAxis']['title'] = dict(text=ylabel)

    if df is not None:
        idx = df.index

        if timestamp_in_idx == 'auto':
            timestamp_in_idx = isinstance(idx[0], datetime.date)

        if timestamp_in_idx:
            res['xAxis']['type'] = 'datetime'

        df = df.applymap(lambda val: dt_to_hc(val) if isinstance(val, (datetime.date, datetime.datetime)) else val)
        if timestamp_in_idx:
            idx = [dt_to_hc(val) for val in idx]
            zip_idx = True

        if zip_idx:
            series.extend(
                dict(name=column, data=zip(idx, list(df[column])))
                for column in df.columns)
        else:
            series.extend(
                dict(name=column, data=list(df[column]))
                for column in df.columns)

    if kwa:
        res = dict_merge(res, kwa, _copy=False)

    return res


def dt_to_hc(dt):
    """ datetime -> highcharts value """
    return int(time.mktime(dt.timetuple()) * 1e3 + dt.microsecond / 1e3)


def ohlc_to_data(df, cols='o h l c'.split(), **kwa):
    """ OHLC dataframe -> data """
    res = [
        ([dt_to_hc(idx.to_datetime())] +
         [row[col] for col in cols])
        for idx, row in df.iterrows()]
    return res


def ohlc_to_serie(df, name=None, extra=None, grouping_units=None, **kwa):
    """ Helper to convert OHLC data to highcharts series """
    if grouping_units is None:
        grouping_units = [
            # unit name, [allowed multiples],
            ['week', [1]],
            ['month', [1, 2, 3, 4, 6]],
        ]
    res_candle = dict(
        type='candlestick',
        name=name,
        # sample data: [[1367366400000,444.46,444.93,434.39,439.29], …]
        data=ohlc_to_data(df, **kwa),
        dataGrouping=dict(units=grouping_units))
    if extra:
        res_candle = dict_merge(res_candle, extra, _copy=False)
    return res_candle


def interleave(joiner, iterable):
    """
    Similar to `str.join` but for lists.

    >>> interleave([1], [2, 3, 4])
    [2, 1, 3, 1, 4]
    """
    try:
        yield next(iterable)
    except StopIteration:  # empty
        return
    for item in iterable:
        yield joiner
        yield item


def ohlcv_to_cdef(df, name, volume='v', pois=None, poi_to_color='default', **kwa):
    """ Helper to convert OHLCV data to a chart def.

    `pois`: [
        # poi
        [
            # point
            [idx, value],
            # point
            [... second point ...]],
        ... more poi ...]
    """
    candle_serie = ohlc_to_serie(df, name, **kwa)
    series = [candle_serie]
    extras = {}
    if volume and volume in df:
        volume_series = df[volume]
        volume_data = [[dt_to_hc(idx.to_datetime()), val]
                       for idx, val in volume_series.iteritems()]
        series.append(dict(
            type='column', name='Volume', data=volume_data,
            yAxis=1,  # Need an axis for that
            dataGrouping=candle_serie['dataGrouping']))
        extras.update(yAxis=[
            dict(labels=dict(align='right', x=-3),
                 title=dict(text='OHLC'),
                 height='60%', lineWidth=2),
            dict(labels=dict(align='right', x=-3),
                 title=dict(text='Volume'),
                 top='65%', height='35%',
                 offset=0, lineWidth=2),
            ])
    if pois:
        breakpointer = [None, None]

        def poi_to_color_default(poi):
            # red if downwards else green
            return 'red' if poi[1][1] < poi[0][1] else 'green'

        if poi_to_color == 'default':
            poi_to_color = poi_to_color_default
        poi_grouped = group((poi_to_color(poi), poi) for poi in pois)
        for color, poi_group in poi_grouped.items():
            poi_data = [
                (dt_to_hc(point_ts) if point_ts else None,
                 point_val)
                for poi in interleave([breakpointer], poi_group)
                for point_ts, point_val in poi]
            series.append(dict(
                type='line', name='POI', data=poi_data,
                color=color,
                # dataGrouping=candle_serie['dataGrouping']
            ))

    return mk_chart_def(series=series, kwa=extras)



# coding: utf8
"""
...
"""

# from __future__ import generator_stop, annotations

import signal
import asyncio
import types
from concurrent.futures import Future


__all__ = (
    '_await',
)


def _await(*args, **kwargs):
    """
    Run an function from a synchronous code, in a new event loop, and await its
    result.

    Intended for debugging asyncs e.g. from ipython.

    Usage:

        result = _await(awaitable, *args, **kwargs)

    Also usable:

        result = _await(awaitable(*args, **kwargs))

    See also:
    https://github.com/django/asgiref/blob/master/asgiref/sync.py
    https://stackoverflow.com/a/48479665
    https://github.com/ipython/ipython/pull/11155
    """
    awaitable, args = args[0], args[1:]

    call_result = Future()
    if isinstance(awaitable, types.CoroutineType):
        assert not args
        assert not kwargs
        cor = awaitable
    else:
        cor = awaitable(*args, **kwargs)

    async def wrap():
        try:
            result = await cor
        except Exception as exc:
            call_result.set_exception(exc)
        else:
            call_result.set_result(result)

    try:
        main_loop = asyncio.get_event_loop()
    except RuntimeError:
        main_loop = None
    if main_loop and main_loop.is_running():
        raise RuntimeError("There's already a running loop, you probably shouldn't use this wrapper")

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(wrap())
    finally:
        try:
            if hasattr(loop, "shutdown_asyncgens"):
                loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            loop.close()
            asyncio.set_event_loop(main_loop)

    return call_result.result()



#!/usr/bin/env python

import os
import sys
import re
import subprocess
import datetime
import logging
import argparse
import shlex
import glob

LOGLEVEL_EVERYTHING = 1
HISTORY_FILE = os.environ.get('HISTORY_FILE', 'HISTORY.rst')
HISTORY_TAG = os.environ.get('HISTORY_TAG', r'(\nRelease History\n-+\n)')

# Note: this is similar to what `bumpversion` does.
HISTORY_VERSION_TAG = os.environ.get(
    'HISTORY_VERSION_TAG',
    # 3 dot-separated numbers or more, ISO datestamp or more
    r'^([0-9]+\.[0-9]+\.[0-9]+[^ ]*) \([0-9]{4}-[0-9]{2}-[0-9]{2}[^)]*\).*')
VERSION_TAG = os.environ.get(
    'VERSION_TAG',
    # Includes 3 groups, the second one should be the current version
    # ans is to be replaced with the new version; without quotes.
    r'''(\n(?:__)?version(?:__)? = ['"])(?P<version>[0-9a-zA-Z.+-]+)(['"] *(?:#.*)?\n)''')
SETUP_VERSION_TAG = os.environ.get('SETUP_VERSION_TAG', VERSION_TAG)
VERSION_FILES = os.environ.get('VERSION_FILES', '*/__init__.py')
VERSION_TAG_TPL = os.environ.get('VERSION_TAG_TPL', '%(version)s')

RELEASE_COMMIT_TPL = os.environ.get('RELEASE_COMMIT_TPL', 'Release %(version)s')


_log = logging.getLogger('_newrelease.py')


def make_parser():
    parser = argparse.ArgumentParser(
        description="Automatically prepare a new pypi release from git")

    parser.add_argument(
        'cmd',
        choices=('prepare', 'check', 'finalise'),
        nargs='?', default='prepare',
        help=(u"Action to perform; usual workflow is 'prepare',"
              u" check `git status`, 'check', then 'finalise'."))

    parser.add_argument(
        '-v', '--verbose',
        dest='verbosity',
        action='count', default=0,
        help="Verbosity level; specify multiple times to increase verbosity")

    parser.add_argument(
        '--major',
        dest='major', action='store_true',
        help="Make a major release (in `prepare`)")
    parser.add_argument(
        '--patch',
        dest='patch', action='store_true',
        help="Make a patch release (third number) (in `prepare`)")
    parser.add_argument(
        '--allow-unclean',
        dest='allow_unclean', action='store_true',
        help="Ignore reposotory taint (in `prepare`)")
    parser.add_argument(
        '--no-commit',
        dest='no_commit', action='store_true',
        help="Do not commit changes (in `finalise`)")

    return parser


def run_sh(*full_command):
    _log.info("Running command: %r", full_command)
    process = subprocess.Popen(
        full_command,
        stdin=subprocess.PIPE,
        shell=False, stdout=subprocess.PIPE)
    stdout, _ = process.communicate()
    _log.debug("Return code: %r", process.returncode)
    assert process.returncode == 0
    return stdout


def run_sh_cmd(cmd, **kwa):
    full_command = shlex.split(cmd)
    return run_sh(*full_command, **kwa)


def _inc_ver_val(val):
    # NOTE: strips non-digit part of the version
    val_digits = re.search('^([0-9]+)', str(val)).group(1)
    return str(int(val_digits) + 1)


def _inc_ver(parts, num):
    """ Increment `num`th part of the version.

    >>> ver_parts = ['1', '2', '3hotfix1']
    >>> _inc_ver(ver_parts, 0)
    ['2', '0', '0']
    >>> _inc_ver(ver_parts, 1)
    ['1', '3', '0']
    >>> _inc_ver(ver_parts, 2)  # non-digit part is stripped
    ['1', '2', '4']
    """
    return (parts[:num] +
            [_inc_ver_val(parts[num])] +
            ['0'] * (len(parts) - 1 - num))


def update_version_in_text(text, new_version, current_version=None):
    match = re.search(VERSION_TAG, text)
    if match is None:
        raise ValueError("Version not found", text)

    pre_tag, current_actual_ver, post_tag = match.groups()
    if current_version is not None:
        if current_version != current_actual_ver:
            _log.error("Found version %r does not match expected %r",
                       current_actual_ver, current_version)

    # pre and post tag parts are supposed to contain the quotes
    new_tag = '%s%s%s' % (pre_tag, new_version, post_tag)
    new_text = '%s%s%s' % (
        text[:match.start()],
        new_tag,
        text[match.end():])
    return new_text


def update_version_in_file(filename, new_version, **kwa):
    with open(filename) as fo:
        text = fo.read()

    try:
        new_text = update_version_in_text(text, new_version, **kwa)
    except ValueError:
        # Re-raise with filename instead of the text
        raise ValueError("Version not found in file %r" % (filename,))

    with open(filename, 'w') as fo:
        fo.write(new_text)

    return new_text


def get_current_version(text=None):
    """ Obtain the current actual version from setup.py.

    Mainly intended for `finalise`.
    """
    if text is None:
        with open('setup.py') as fo:
            text = fo.read()

    match = re.search(VERSION_TAG, text)
    if not match:
        raise ValueError("Could not find version in setup.py")

    return match.group(2)


def get_git_mods():
    return run_sh_cmd('git status --porcelain')


def get_git_taint():
    return run_sh_cmd('git clean -d -x -n')


def prepare(params):
    if not params.allow_unclean:
        mods = get_git_mods()
        if mods.strip():
            _log.critical("Cannot proceed as git repo has modifications:\n%r", mods)
            return 14
        taint = get_git_taint()
        if taint.strip():
            _log.critical("Cannot proceed as git repo is not clean:\n%r", taint)
            return 13

    with open(HISTORY_FILE) as fo:
        history = fo.read()

    history_tag = HISTORY_TAG
    history_parts = re.split(history_tag, history, flags=re.MULTILINE)
    _log.debug("History parts: %r", [val[:10] for val in history_parts])
    history_versions = history_parts[-1]

    current_version_match = re.search(
        HISTORY_VERSION_TAG,
        history_versions, re.MULTILINE)
    current_version = current_version_match.group(1)
    today = datetime.date.today().isoformat()
    version_parts = current_version.split('.')

    new_version = os.environ.get('NEW_VERSION')
    if not new_version:
        if params.major:
            version_parts = _inc_ver(version_parts, 0)
        elif params.patch:
            version_parts = _inc_ver(version_parts, 2)
        else:
            version_parts = _inc_ver(version_parts, 1)

        new_version = '.'.join(version_parts)

    _log.debug("New version: %s", new_version)

    version_tag = VERSION_TAG_TPL % dict(version=current_version)
    git_history = run_sh(
        "git", "log", "%s..HEAD" % (version_tag,),
        "--format=format: - %s")
    if isinstance(git_history, bytes):
        git_history = git_history.decode('utf-8', errors='replace')
    _log.debug("Git history: %r", git_history)

    new_history_header = "%s (%s)" % (new_version, today)
    new_history_header_full = "%s\n%s" % (
        new_history_header,
        '+' * len(new_history_header))
    new_history_versions = "\n%s\n\n%s\n\n%s" % (
        new_history_header_full,
        git_history,
        history_versions)
    new_history_full = ''.join(
        history_parts[:-1] +
        [new_history_versions])
    _log.log(LOGLEVEL_EVERYTHING, "New history: \n%s", new_history_full)

    with open(HISTORY_FILE, 'w') as fo:
        fo.write(new_history_full)

    try:
        update_version_in_file('setup.py', new_version,
                               current_version=current_version)
        for filename in glob.glob(VERSION_FILES):
            update_version_in_file(filename, new_version,
                                   current_version=current_version)
    except ValueError as exc:
        _log.critical("Error updating version: %r", exc)
        return 2

    _log.info("Done")


def check(params):
    run_sh_cmd('python setup.py develop')
    run_sh_cmd('python setup.py test')
    if os.path.exists('test.sh'):
        run_sh('./test.sh')


def finalise(params):
    try:
        version = get_current_version()
    except ValueError as exc:
        _log.critical("Error getting current version: %r", exc)
        return 2

    _log.info("Found current version: %r", version)

    if not params.no_commit:
        mods = get_git_mods()
        if not mods.strip():
            _log.critical((
                "Finalise should be called after `prepare` made the"
                " changes and while they are not committed yet; alas, no"
                " changes are found by git"))
            return 4

    taint = get_git_taint()
    if taint.strip():
        _log.critical((
            "There are git-untracked files which should not exist;"
            " please run `git clean -d -x -f` to remove them before"
            " using `finalise`:\n%s"), taint)
        return 5

    tpl_env = dict(version=version)

    release_msg = RELEASE_COMMIT_TPL % tpl_env
    if not params.no_commit:
        run_sh('git', 'commit', '-a', '-m', release_msg)

    version_tag = VERSION_TAG_TPL % tpl_env
    run_sh('git', 'tag', '-a', version_tag, '-m', release_msg)
    run_sh_cmd('git push')
    run_sh_cmd('git push --tags')

    run_sh_cmd('python setup.py sdist')
    run_sh_cmd('python setup.py bdist_wheel')
    run_sh_cmd('twine upload ./dist/*')

    _log.debug("Done.")


def main(args=None):
    parser = make_parser()
    params = parser.parse_args(args)

    _log_levels = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
        3: LOGLEVEL_EVERYTHING,
    }
    logging.basicConfig(level=_log_levels[min(params.verbosity, 3)])

    if params.cmd == 'prepare':
        return prepare(params)
    elif params.cmd == 'finalise':
        return finalise(params)


if __name__ == '__main__':
    sys.exit(main())



# coding: utf8
""" madstuff: datadiff stuff """

from __future__ import print_function, unicode_literals, absolute_import, division

import re
import sys
import difflib
import itertools

import yaml

from .base import colorize_yaml, colorize_diff


__all__ = (
    '_dumprepr',
    '_diff_pre_diff', '_diff_datadiff_data', 'datadiff', 'p_datadiff',
)


def _dumprepr(val, no_anchors=True, colorize=False, try_ujson=True, allow_unsorted_dicts=False, **kwa):
    """ Advanced-ish representation of an object (using YAML) """
    dumper = yaml.SafeDumper

    # NOTE: this means it'll except on infinitely-recursive data.
    if no_anchors:

        class NoAliasesSafeDumper(dumper):

            def ignore_aliases(self, *args, **kwargs):  # pylint: disable=arguments-differ
                return True

        dumper = NoAliasesSafeDumper

    if allow_unsorted_dicts and sys.version_info >= (3, 7):
        from yaml.nodes import MappingNode, ScalarNode

        class UnsortingDumper(dumper):
            """ ... """

            def represent_mapping(self, tag, mapping, flow_style=None):
                value = []
                node = MappingNode(tag, value, flow_style=flow_style)
                if self.alias_key is not None:
                    self.represented_objects[self.alias_key] = node
                best_style = True
                if hasattr(mapping, 'items'):
                    mapping = list(mapping.items())
                for item_key, item_value in mapping:
                    node_key = self.represent_data(item_key)
                    node_value = self.represent_data(item_value)
                    if not (isinstance(node_key, ScalarNode) and not node_key.style):
                        best_style = False
                    if not (isinstance(node_value, ScalarNode) and not node_value.style):
                        best_style = False
                    value.append((node_key, node_value))
                if flow_style is None:
                    if self.default_flow_style is not None:
                        node.flow_style = self.default_flow_style
                    else:
                        node.flow_style = best_style
                return node

        dumper = UnsortingDumper

    params = dict(
        # Convenient upper-level kwarg for the most often overridden thing:
        default_flow_style=kwa.pop('default_flow_style', False),
        allow_unicode=True,
        encoding=None,  # return text
        Dumper=dumper,
    )
    params.update(kwa.get('yaml_kwa', {}))

    def maybe_colorize(text):
        if not colorize:
            return text
        return colorize_yaml(text, **kwa)

    res = ''
    try:
        res += maybe_colorize(yaml.dump(val, **params))
    except Exception as exc:
        if not try_ujson:
            raise
        # ujson can handle many objects somewhat-successfully. But can
        # segfault while doing that.
        import ujson
        res += "# Unable to serialize directly! (%r)\n" % (exc,)
        prepared_value = ujson.loads(ujson.dumps(val))  # pylint: disable=c-extension-no-member
        res += maybe_colorize(yaml.dump(prepared_value, **params))

    return res


def _diff_pre_diff(val, **kwa):
    """ Prepare a value for diff-ing """
    _repr = kwa.get('_repr', _dumprepr)
    res = _repr(val, **kwa)
    res = res.splitlines()
    return res


def word_diff_color(val1, val2, add='\x1b[32m', rem='\x1b[31;01m',
                    clear='\x1b[39;49;00m', n=3):
    """ Proper-ish word-diff represented by colors """

    def _preprocess(val):
        return re.split(r'(?u)(\w+)', val)

    diffs = difflib.unified_diff(_preprocess(val1), _preprocess(val2), n=n)

    def _postprocess(line):
        if line in ('--- \n', '+++ \n'):
            return ''
        if line.startswith('+'):
            color = add
        elif line.startswith('-'):
            color = rem
        else:
            color = ''
        return '%s%s%s' % (color, line[1:], clear)

    diffs = list(diffs)
    diffs_colored = (_postprocess(line) for line in diffs)
    # return diffs_colored
    print(''.join(diffs_colored))


def _diff_datadiff_data(val1, val2, n=3, **kwa):
    """ Do the diff and return the data """
    val1_p = _diff_pre_diff(val1, **kwa)
    val2_p = _diff_pre_diff(val2, **kwa)
    res = difflib.unified_diff(val1_p, val2_p, n=n)
    return res


def datadiff(val1, val2, colorize=False, colorize_as_yaml=False, **kwargs):
    """ Return a values diff string """
    kwargs['colorize'] = colorize_as_yaml  # NOTE: controversial
    data = _diff_datadiff_data(val1, val2, **kwargs)

    # line_limit
    _ll = kwargs.pop('line_limit', 200)
    if _ll:
        data_base = data
        data = list(itertools.islice(data_base, _ll))
        try:
            next(data_base)
        except StopIteration:
            pass
        else:
            data.append('...')  # u'…'

    res = '\n'.join(data)
    if colorize:
        res = colorize_diff(res, **kwargs)
    return res


def p_datadiff(val1, val2, **kwargs):
    """ Print the values diff """
    # TODO: yaml coloring *and* diff coloring?
    print(datadiff(val1, val2, **kwargs))



# coding: utf8
""" Various methods and helpers for code profiling (especially
LineProfiler)

Example usage of this all:
  * `
    prof.wrap_packages(['mypackage.somemodule'])
    try:
        run_it_all()
    finally:
        prof.profile.dump_stats('some_file.prof')
    `
  * `python -m line_profiler some_file.prof`
  * Bang you head into the output
It is also possible to use `get_prof()` before all other imports and add
 `@profile` wrapper specifically to the target functions, instead of
 using the `wrap_packages`; it is also possible to use `kernprof.py`
 instead of `get_prof` and this whole module.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import sys
import inspect
import traceback


class DummyProfiler(object):
    """ Dummy LineProfiler-like wrapper """
    def __call__(self, fn, *ar, **kwa):
        return fn  # do nothing as a wrapper
    def __enter__(self):
        return self  # do nothing as a context
    ## Do nothing for other LineProfiler-like functions
    __exit__ = enable = disable = dump_stats = print_stats = lambda self, *ar, **kwa: None


profile = None
profile_type = None


def _check_add_builtin(force=True):
    ## Do the scary global-state stuff
    global profile


def get_prof(proftype='lp', add_builtin=True):
    """ Initialize the global and buitin profiler *wrapper* (currently,
    only meant for LineProfiler) """
    global profile
    global profile_type
    if profile is not None and profile_type == proftype:
        pass  ## ... nothing to generate.
    ## Otherwise - make one.
    elif proftype == 'lp':
        import line_profiler
        profile = line_profiler.LineProfiler()
    elif proftype == 'dummy':
        profile = DummyProfiler()
    else:
        raise Exception("Unknown `proftype`.")
    if add_builtin:  # add/replace/whatever
        from six.moves import builtins
        builtins.__dict__['profile'] = profile
    return profile


def _wrap_class_stuff(the_class, wrapf):
    # for i_name in dir(the_class):
    #     i_val = getattr(the_class, i_name)
    #     if not callable(i_val):
    #         continue
    # # Note: this, still, executes the class properties; hopefully, that
    # #   won't be a problem (they are damn rare after all)
    for i_name, i_val in inspect.getmembers(the_class, predicate=inspect.ismethod):
        try:
            if getattr(i_val, '__prof_wrapped__', False):
                continue  # do not wrap wrapped wrapstuff.
            # XXX: do the sourcefilename check too?
            print("    Method wrapping: %s" % (i_name,))
            wrapped_val = wrapf(i_val)
            setattr(wrapped_val, '__prof_wrapped__', True)
            setattr(the_class, i_name, wrapped_val)
        except Exception as e:
            print("     ... Failed (%r). %r" % (i_name, e))


def _wrap_module_stuff(module, wrapf):
    """ Wrap each function and method in the module with the specified
    function (e.g. LineProfiler) """
    # This should not fail:
    # (Note: there's also `inspect.getfile; but it sometimes returns ...py and sometimes ...pyc)
    if not inspect.ismodule(module) and isinstance(module, object):
        # NOTE: only known to happen with twisted.internet.reactor;
        # along with other problematic twistedness.
        module = type(module)
    module_filename = inspect.getsourcefile(module)
    # ...
    for i_name in dir(module):
        try:
            i_val = getattr(module, i_name)
            if getattr(i_val, '__prof_wrapped__', False):
                continue  # do not wrap wrapped wrapstuff.
            try:
                i_filename = inspect.getsourcefile(i_val)
            except Exception as e:
                i_filename = None
            if i_filename is not None and i_filename != module_filename:
                ## Skip stuff that we know isn't from that module
                continue
            elif i_filename is None:
                ## ... and skip, too, all the built-ins, non-(module/class/function/...)
                continue
            if inspect.isclass(i_val):
                print("  Class wrapping: %s" % (i_name,))
                _wrap_class_stuff(i_val, wrapf=wrapf)
            elif callable(i_val):
                print("  Wrapping: %s" % (i_name,))
                wrapped_val = wrapf(i_val)
                setattr(wrapped_val, '__prof_wrapped__', True)
                setattr(module, i_name, wrapped_val)
        except Exception as e:
            print("  ... Failed (%r). %r" % (i_name, e))


# TODO: make it possible to specify down to particular objects to wrap.
def wrap_packages(packages, wrapf=None, verbose=True):
    """ Wraps all the currently imported modules within any package of
    the `packages` (e.g. `['module1', 'package2', 'package3.module1']`) """
    # Line-profiler: package-wrap.
    # NOTE: this would be more reliable if done on the import hook; but
    #   this is more simple and should be sufficient here.
    if wrapf is None:
        wrapf = get_prof()
    for pkg_name, pkg in sys.modules.items():
        if pkg is None:
            continue
        for pkg_to_wrap in packages:
            if pkg_name.startswith(pkg_to_wrap):
                if verbose:
                    print("Module-wrapping: %s" % (pkg_name,))
                try:
                    _wrap_module_stuff(pkg, wrapf)
                except Exception as exc:
                    print(" ... Failed.", exc)
                break  # make sure not to wrap it many times.
    if verbose:
        print("Wrapping done.")


## TODO?: put into the main prof module
def stgrab(sig, frame):
    """ function that is supposed to be addd as a signal handler
    (e.g. on USR2) and prints the stack trace when called """
    ofra = inspect.getouterframes(frame)
    d = dict(_frame=frame)
    d.update(frame.f_globals)
    d.update(frame.f_locals)
    trace_data = [(v[1], v[2], v[3], ''.join(v[4] or [])) for v in reversed(ofra)]
    trace_formatted = ''.join(traceback.format_list(trace_data))
    trace_last_src = ''.join(ofra[0][4] or [])
    if (stgrab.check_polling
            and any((v in trace_last_src) for v in stgrab.poll_codes)):
        print(" ... Polling ...")  # don't spam that specific traceback
        return
    print(
        " ------- %s\n" % (stgrab.header_str,) +
        # "Framedata: %s\n" % (name, d) +
        "Traceback:\n%s" % (trace_formatted,) +
        "")


## Funny place to put that:
stgrab.header_str = "Stack trace in process:"
stgrab.check_polling = True
## Source code pieces of the known lines that do polling.
## MAYBEDO: add other known polling lines
stgrab.poll_codes = ['   l = self._poller.poll(timeout']



# coding: utf-8
"""
Logging: annotating filters

https://docs.python.org/2/howto/logging-cookbook.html#adding-contextual-information-to-your-logging-output
"""

import time
import logging
import socket  # for hostnames
from .base import simple_memoize_argless


__all__ = (
    'time_diff_annotator',
    'short_hostname_annotator',
    'full_hostname_annotator',
)


_not_available = object()


class Annotator(logging.Filter):
    """ A convenience abstract class for most annotators """

    attribute_name = None
    # Shortcut for using the previously generated value.
    use_cached_value = True

    def __init__(self, *args, **kwargs):
        self.attribute_name = kwargs.pop('attribute_name', None) or self.attribute_name
        if self.attribute_name is None:
            raise Exception(
                "attribute_name should either be on class or always specified")
        super(Annotator, self).__init__(*args, **kwargs)

    def get_value(self, record, *args, **kwargs):
        raise NotImplementedError

    def _get_value_cached(self, record, *args, **kwargs):
        value = getattr(record, self.attribute_name, _not_available)
        if value is not _not_available:
            return value
        return self.get_value(record, *args, **kwargs)

    def filter(self, record):
        """ “annotate”, actually """
        if self.use_cached_value:
            value = self._get_value_cached(record)
        else:
            value = self.get_value(record)
        setattr(record, self.attribute_name, value)
        return True


class time_diff_annotator(Annotator):
    """ A simple filter that adds `time_diff` to the record, which
    shows the time from the last log line of the same process. Mostly
    useful in development. """

    attribute_name = 'time_diff'

    def __init__(self, *args, **kwargs):
        self.last_ts = time.time()
        super(time_diff_annotator, self).__init__(*args, **kwargs)

    def get_value(self, record, *args, **kwargs):
        now = time.time()
        result = now - self.last_ts
        self.last_ts = now
        return result


def make_simple_annotating_filter(name, func):

    def get_value(self, *args, **kwargs):
        return func()

    filter_class = type(
        '_%s_annotator', (Annotator,),
        dict(attribute_name=name, get_value=get_value))
    return filter_class


# getfqdn occasionally might use network, which is why it is better to
# cache it.
cached_getfqdn = simple_memoize_argless(socket.getfqdn)


short_hostname_annotator = make_simple_annotating_filter('hostname', socket.gethostname)
full_hostname_annotator = make_simple_annotating_filter('hostname', cached_getfqdn)


def get_celery_task_attributes():
    result = dict(task_name=None, task_id=None, meta=None, meta_meta=None)
    try:
        # See celery.app.log.TaskFormatter
        from celery._state import get_current_task
        task = get_current_task()
        if not task:
            return dict(result, meta='no_task')
        if not task.request:
            return dict(result, meta='no_task_request')
        return dict(result, task_name=task.name, task_id=task.request.id)
    except Exception as exc:
        return dict(result, meta='error', meta_meta=exc)


# NOTE: using these together could be more performant, but at the moment too bothersome.
celery_task_name_annotator = make_simple_annotating_filter(
    'celery_task_name', lambda: get_celery_task_attributes()['task_name'])
celery_task_id_annotator = make_simple_annotating_filter(
    'celery_task_id', lambda: get_celery_task_attributes()['task_id'])


class celery_process_name_annotator(Annotator):

    attribute_name = 'celery_process'
    skip_main_process = True

    def get_value(self, *args, **kwargs):
        # see celery.utils.log
        try:
            from billiard import current_process
            result = current_process()._name
            if self.skip_main_process and result == 'MainProcess':
                return
            return result
        except Exception:
            return



# coding: utf8
""" madstuff: oneliners and debug-useful stuff """

from __future__ import print_function, unicode_literals, absolute_import, division

import sys
import traceback
import six
from .base import o_repr
from .madness_reprstuff import GenReprWrapWrap
from .madness_datadiff import _dumprepr


__all__ = (
    '_try', '_try2',
    '_iter_ar', '_filter', '_filter_n',
    '_ipdbg', '_ipdbt',
    '_print', '_uprint', '_yprint', '_mrosources', 'p_o_repr',
)


def _try2(_function_thingie, *ar, **kwa):
    """ Returns (res, None) on success or (None, exception) """
    # The weird names are to minimise kwa collision
    _exc_clss = kwa.pop('_exc_clss', Exception)
    # TODO?: return namedtuple?
    try:
        return (_function_thingie(*ar, **kwa), None)
    except _exc_clss as e:
        return (None, e)


def _try(*ar, **kwa):
    """ Return call result or None if an exception occurs """
    return _try2(*ar, **kwa)[0]


def _iter_ar(*args):
    """ Helper to get an iterable (mostly tuple) out of some arguments """
    if not args:
        raise TypeError("At least one argument was required")
    if len(args) > 1:
        return args

    first_arg = args[0]
    if hasattr(first_arg, '__iter__'):
        return first_arg
    return args


@GenReprWrapWrap
def _filter(*ar):
    """ Mostly the same as `filter(None, …)` but as a generator with
    conveniences. """
    it = _iter_ar(*ar)
    for i in it:
        if i:
            yield i


@GenReprWrapWrap
def _filter_n(*ar):
    """ Filter out None specifically (also a generator with conveniences) """
    it = _iter_ar(*ar)
    for i in it:
        if i is not None:
            yield i


def _print(something):
    """ Simple one-argument `print` one-liner; returns the argument """
    print(something)
    return something


def _ipdbg(_a_function_thingie, *ar, **kwa):
    """ Run with ipdb post-mortem on exception """
    import ipdb
    try:
        return _a_function_thingie(*ar, **kwa)
    except Exception as exc:
        _, _, sys.last_traceback = sys.exc_info()
        traceback.print_exc()
        ipdb.pm()
        return None


def _ipdbt(_a_function_thingie, *ar, **kwa):
    """ Run with ipdb trace and post-mortem on exception """
    import ipdb
    ipdb.set_trace()
    try:
        return _a_function_thingie(*ar, **kwa)
    except Exception as exc:
        _, _, sys.last_traceback = sys.exc_info()
        traceback.print_exc()
        ipdb.pm()
        return None


def _pdbg(_a_function_thingie, *ar, **kwa):
    """ Run with pdb post-mortem on exception """
    import pdb
    try:
        return _a_function_thingie(*ar, **kwa)
    except Exception as exc:
        _, _, sys.last_traceback = sys.exc_info()
        traceback.print_exc()
        pdb.pm()
        return None


def _pdbt(_a_function_thingie, *ar, **kwa):
    """ Run with pdb trace and post-mortem on exception """
    import pdb
    pdb.set_trace()
    try:
        return _a_function_thingie(*ar, **kwa)
    except Exception as exc:
        _, _, sys.last_traceback = sys.exc_info()
        traceback.print_exc()
        pdb.pm()
        return None


def _uprint(obj, ret=False):
    try:
        from IPython.lib.pretty import pretty
    except Exception:
        from pprint import pformat as pretty
    obj_repr = pretty(obj)
    if isinstance(obj_repr, bytes):  # py2
        obj_repr = obj_repr.decode('unicode-escape')
    print(obj_repr)
    if ret:
        return obj
    return None


def _yprint(obj, ret=False, **kwa):
    kwa.setdefault('colorize', True)
    kwa.setdefault('no_anchors', False)
    kwa.setdefault('default_flow_style', None)
    kwa.setdefault('allow_unsorted_dicts', True)
    res_text = _dumprepr(obj, **kwa)
    if six.PY2 and isinstance(res_text, six.text_type):
        res_text = res_text.encode("utf-8")
    print(res_text)
    if ret:
        return obj
    return None


def _mrosources(cls, attname, raw=False, colorize=False):
    """ Return a (prettified) source code of a class method in its
    MRO. For figuring out where does the super() actually go. """
    import inspect
    from pyaux import dotdict
    meths = [dotdict(cls=mrocls, meth=getattr(mrocls, attname, None))
             for mrocls in cls.__mro__]
    meths = [val for val in meths if val.meth]

    def getsrc(val):
        try:
            return inspect.getsource(val)
        except Exception:
            return '<???>'

    meths = [dotdict(val, src=getsrc(val.meth))
             for val in meths]

    if colorize:
        from pyaux.base import colorize as colorfunc
        colorfunc_kwa = dict(fmt='py')
        if isinstance(colorize, dict):
            colorfunc_kwa.update(colorize)
        meths = [dotdict(val, src=colorfunc(val.src, **colorfunc_kwa))
                 for val in meths]
    if raw:
        return meths
    res = '\n\n'.join(
        ' ======= %s =======\n%s' % (
            val.cls.__name__, val.src)
        for val in meths)
    return res


def p_o_repr(o, **kwa):
    kwa = dict(dict(_colors=True, _colors256=True), **kwa)
    print(o_repr(o, **kwa))



"""
Entry-module for the pyx-compiled `_datadeque` module.

WARNING: this module will be moved to `pyauxm`.
"""

try:
    from . import _datadeque
except ImportError as e1:
    #try:
    import pyximport
    ## XX: problematic: this will: 1. install pyximport for everything
    ##   (can be unintended), 2. try to compile the module on import,
    ##   which can take a while and fail with an error too. For that
    ##   reason, it is not done in the main module.
    pyximport.install()
    from . import _datadeque
    #except ImportError as e2:
    #    warnings.warn("Error importing pyx (Cython) modules: %r" % (e1,))
from ._datadeque import *



#!/usr/bin/env python
""" lzcat """

from __future__ import print_function, unicode_literals, absolute_import, division

import sys
import pylzma
from pyaux.lzmah import get_stdin, get_stdout


def unlzma(fi, fo, fi_close=True, fo_close=True, bufs=6553500):
    """ Decompress `fi` into `fo` (`file` or filename) """
    if isinstance(fi, str):
        fi, fi_n = open(fi, 'rb'), fi
        # fi_close = True
    if isinstance(fo, str):
        fo, fo_n = open(fo, 'wb'), fo
        # fo_close = True
    # i.seek(0)

    # XXXX: better way?
    #  * s.decompress *requires* an `output buffer size`, i.e. size of the
    #   unpacked data, otherwise packed data is stored in internal buffers
    #   and returned on flush (which gets problematic).
    #  * Suggested default is to read by 1 byte and use the default buffer
    #   size.  Which gets slow.
    #  * Nearest hax to fix is to use output buffer over 30x (or something)
    #   the size of input buffer.  Which is not a nice thing to do, but
    #   works...  mostly.
    #  * ... anyway, symptoms: slowdown on decompression down to zero speed,
    #   high memory usage (up to almost size of the uncompressed file),
    #   after which all the decompressed data is written in one go.
    in_bufs = int(bufs / 100)
    s = pylzma.decompressobj()
    while True:
        tmp = fi.read(in_bufs)
        if not tmp:
            break
        fo.write(s.decompress(tmp, bufs))
    fo.write(s.flush())
    if fo_close:
        fo.close()
    if fi_close:
        fi.close()
    return fi, fo


def _lzcat_main():
    if len(sys.argv) > 1:
        fi_a = sys.argv[1]
    else:
        fi_a = '-'
    if len(sys.argv) == 3:
        fo_a = sys.argv[2]
    else:
        fo_a = '-'
    if len(sys.argv) > 3:
        print("Basic usage: %s [<from_file> [<to_file>]]" % (sys.argv[0],))
        sys.exit()

    if fi_a == '-':
        fi_a = get_stdin()
    if fo_a == '-':
        fo_a = get_stdout()
    unlzma(fi_a, fo_a)


if __name__ == '__main__':
    _lzcat_main()



# coding: utf8
"""
The most simple and backwards-compatible way of getting something. Not
necessarily performant.

Does not contain unsafe-to-import modules.
"""

# Ordering is from least important to most important, with modules
# themselves last.

from .madness import *
from .runlib import *
from .interpolate import *
from .dicts import *
from .urlhelpers import *
from .ranges import *
from .base import *
from . import madness
from . import runlib
from . import interpolate
from . import dicts
from . import urlhelpers
from . import ranges
from . import base
